import numpy as np
import networkx as nx
import torch
import os
from tqdm import tqdm
import matplotlib.pyplot as plt
from torch_geometric.data import Data, InMemoryDataset
from torch_geometric.utils import erdos_renyi_graph

def vis_from_pyg(data, filename = None):
    edges = data.edge_index.T.cpu().numpy()
    labels = data.x[:,0].cpu().numpy()

    g = nx.Graph()
    g.add_edges_from(edges)

    # dropped_nodes = np.ones(labels.shape[0]).astype(bool)
    for ilabel in range(labels.shape[0]):
        if ilabel not in np.unique(edges):
            g.add_node(ilabel)
    # labels = labels[dropped_nodes]

    fig, ax = plt.subplots(figsize = (6,6))

    pos = nx.kamada_kawai_layout(g)

    nx.draw_networkx_edges(g, pos = pos, ax = ax)
    nx.draw_networkx_nodes(g, pos = pos, node_color=labels, cmap="tab20",
                           vmin = 0, vmax = 20, ax = ax)

    ax.axis('off')

    plt.tight_layout()
    if filename is None:
        plt.show()
    else:
        plt.savefig(filename)
        plt.close()


def get_random_graph():

    size = np.random.randint(low = 12, high = 128)

    rho = 0.05 + 0.25 * np.random.random()

    edge_index = erdos_renyi_graph(size, rho)

    G = Data(edge_index = edge_index)

    return G, rho

def get_random_dataset(num = 1000):

    nx_graph_list_rhos = [get_random_graph() for _ in tqdm(range(num), leave=False)]
    datalist = [item[0] for item in nx_graph_list_rhos]
    rhos= [item[1] for item in nx_graph_list_rhos]

    for idata, data in enumerate(datalist):
        data.y = torch.Tensor([rhos[idata]])
        datalist[idata] = data


    return datalist





class RandomDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`

    Dataset of random erdos-renyi graphs, between 12 and 128 nodes, of random density between 0.05 and 0.3.

    The target is the density of each graph.

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. -1 takes all available samples for that stage. (default: :obj:`-1`).
    """
    def __init__(self, root, stage="train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return []


    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Random files exist")
            return

        data_list = get_random_dataset(num=self.num)

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        # if self.stage != "train":
        #     for i, data in enumerate(data_list):
        #         vis_from_pyg(data, filename=self.root + f'/processed/{self.stage}-{i}.png')

        for data in data_list:
            print(data)

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


import os
import torch
from torch_geometric.data import InMemoryDataset, Data
from tqdm import tqdm
from torch.utils.data.dataset import ConcatDataset


class ToPDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    Contributor email: `alexander.davies@bristol.ac.uk`
    
    Processes an InMemoryDataset into a ToP dataset by removing node and edge features.
    
    Based on the paper:

         `Towards Generalised Pre-Training of Graph Models, Davies, A. O., Green, R. W., Ajmeri, N. S., and Silva Filho, T. M.,  arXiv e-prints, 2024. doi:10.48550/arXiv.2311.03976.`

    The resulting dataset is topology-only, intended for pre-training with ToP, and as such this module does not produce validation/test splits.

    Saves a single processed file `train-top.pt` under `./root/processed/`.

    Args:
        original_dataset (InMemoryDataset): The original dataset to convert to ToP format.
        root (str): Root directory where the dataset should be saved. If 'none', will use the root directory of `original_dataset`. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. `num=-1` will convert all available samples from the original. (default: :obj:`-1`).
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)

    """

    def __init__(self, original_dataset, stage = "train", root = None,  num = -1, transform=None, pre_transform=None, pre_filter=None):
        self.original_dataset = original_dataset
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        self.num = num
        root = root if root is not None else original_dataset.root
        print(f"\n\nConverting original dataset {original_dataset} at {root}")
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return ['dummy.csv']

    @property
    def processed_file_names(self):
        return ['train-top.pt',
                'val-top.pt',
                'test-top.pt']


    def process(self):
        # Read data into huge `Data` list.
        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print(f"Processed ToP files exist at {self.processed_paths[self.stage_to_index[self.stage]]}")
            return
        
        data_list = self.original_dataset# get_fb_dataset(num=self.num)
        new_data_list = []
        if type(data_list) == ConcatDataset:
            print("Found concat dataset in ToP conversion")
            for dataset in data_list.datasets:
                for item in dataset:
                    new_data_list.append(item)
            self.original_dataset = new_data_list
            data_list = new_data_list

        num_samples = len(data_list)
        if self.num == -1 or num_samples < self.num:
            keep_n = num_samples
        else:
            keep_n = self.num

        print(f"Dropping targets from {keep_n} samples out of {num_samples}")
        new_data_list = []
        for i, item in enumerate(tqdm(data_list[:keep_n], desc = "Converting dataset to ToP")):

            if item.x is not None:
                n_nodes = item.x.shape[0]
            else:
                n_nodes = torch.max(item.edge_index) + 1

            n_edges =  item.edge_index.shape[1]
            data = Data(x = torch.ones(n_nodes).to(torch.int).reshape((-1, 1)),
                        edge_index=item.edge_index,
                        edge_attr=torch.ones(n_edges).to(torch.int).reshape((-1,1)),
                        y = None if self.stage == "train" else item.y)

            new_data_list.append(data)
        data_list = new_data_list

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])



from typing import Callable, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.typing import OptPairTensor, Adj, OptTensor, Size
from torch_sparse import SparseTensor

from unsupervised.convs.inits import reset


class GINEConv(MessagePassing):
    def __init__(self, nn: Callable, eps: float = 0., train_eps: bool = False,
                 **kwargs):
        kwargs.setdefault('aggr', 'add')
        super(GINEConv, self).__init__(**kwargs)
        self.nn = nn
        self.initial_eps = eps
        if train_eps:
            self.eps = torch.nn.Parameter(torch.Tensor([eps]))
        else:
            self.register_buffer('eps', torch.Tensor([eps]))
        self.reset_parameters()

    def reset_parameters(self):
        reset(self.nn)
        self.eps.data.fill_(self.initial_eps)

    def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj,
                edge_attr: OptTensor = None, edge_weight: OptTensor = None, size: Size = None) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # Node and edge feature dimensionalites need to match.
        if isinstance(edge_index, Tensor):
            assert edge_attr is not None
            assert x[0].size(-1) == edge_attr.size(-1)
        elif isinstance(edge_index, SparseTensor):
            assert x[0].size(-1) == edge_index.size(-1)
        
        vals, counts = torch.unique(edge_weight, return_counts = True)

        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, edge_weight=edge_weight, size=size)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    def message(self, x_j: Tensor, edge_attr: Tensor, edge_weight) -> Tensor:
        return F.relu(x_j + edge_attr) if edge_weight is None else F.relu(x_j + edge_attr) * edge_weight.view(-1, 1)

    def __repr__(self):
        return '{}(nn={})'.format(self.__class__.__name__, self.nn)


import copy

import numpy as np
import torch
from torch_geometric.data import Data


def initialize_edge_weight(data):
	data.edge_weight = torch.ones(data.edge_index.shape[1], dtype=torch.float)
	return data

def initialize_node_features(data):
	num_nodes = int(data.edge_index.max()) + 1
	data.x = torch.ones((num_nodes, 1))
	return data

def set_tu_dataset_y_shape(data):
	num_tasks = 1
	data.y = data.y.unsqueeze(num_tasks)
	return data

def drop_nodes(data, ratio=0.1):
	if hasattr(data, "x"):
		node_num, _ = data.x.size()
	else:
		node_num = int(data.edge_index.max()) + 1

	_, edge_num = data.edge_index.size()
	drop_num = int(node_num * ratio)
	idx_drop = np.random.choice(node_num, drop_num, replace=False)
	idx_nondrop = [n for n in range(node_num) if not n in idx_drop]
	idx_dict = {idx_nondrop[n]: n for n in list(range(node_num - drop_num))}

	x_aug = data.x[idx_nondrop]

	remove_edge_mask = torch.zeros(edge_num, dtype=torch.bool)

	for n_idx in idx_drop:
		temp_mask = torch.logical_or(data.edge_index[0] == n_idx, data.edge_index[1] == n_idx)
		remove_edge_mask = torch.logical_or(temp_mask, remove_edge_mask)

	edge_index_aug = data.edge_index[:, ~remove_edge_mask]

	edge_num_aug = edge_index_aug.shape[1]
	edge_index_aug = edge_index_aug.numpy()
	edge_index_aug = [[idx_dict[edge_index_aug[0, n]], idx_dict[edge_index_aug[1, n]]] for n in range(edge_num_aug) if
	                  not edge_index_aug[0, n] == edge_index_aug[1, n]]
	try:
		edge_index_aug = torch.tensor(edge_index_aug).transpose_(0, 1)
	except:
		#         print(data)
		#         print(data.edge_index)
		#         print(idx_drop)
		#         print(idx_nondrop)
		#         print(idx_dict)
		#         print(~remove_edge_mask)
		#         print(edge_index_aug)
		#         print("Removal of all edges happened ... Returning original data")
		edge_index_aug = torch.tensor([[], []], dtype=torch.long)
	#         data_aug = Data(x=data.x, edge_index=data.edge_index, edge_attr=data.edge_attr)
	#         return data_aug

	if hasattr(data, "edge_attr")  and data.edge_attr is not None:
		edge_attr_aug = data.edge_attr[~remove_edge_mask]
		data_aug = Data(x=x_aug, edge_index=edge_index_aug, edge_attr=edge_attr_aug)
	else:
		data_aug = Data(x=x_aug, edge_index=edge_index_aug)

	return data_aug


def subgraph(data, ratio=0.2):
	node_num, _ = data.x.size()
	_, edge_num = data.edge_index.size()
	sub_num = int(node_num * ratio)

	edge_index = data.edge_index.numpy()

	idx_sub = [np.random.randint(node_num, size=1)[0]]
	idx_neigh = set([n for n in edge_index[1][edge_index[0] == idx_sub[0]]])

	count = 0
	while len(idx_sub) <= sub_num:
		count = count + 1
		if count > node_num:
			break
		if len(idx_neigh) == 0:
			break
		sample_node = np.random.choice(list(idx_neigh))
		if sample_node in idx_sub:
			continue
		idx_sub.append(sample_node)
		idx_neigh.union(set([n for n in edge_index[1][edge_index[0] == idx_sub[-1]]]))

	idx_drop = [n for n in range(node_num) if not n in idx_sub]
	idx_nondrop = idx_sub
	idx_dict = {idx_nondrop[n]: n for n in list(range(len(idx_nondrop)))}

	x_aug = data.x[idx_nondrop]

	remove_edge_mask = torch.zeros(edge_num, dtype=torch.bool)

	for n_idx in idx_drop:
		temp_mask = torch.logical_or(data.edge_index[0] == n_idx, data.edge_index[1] == n_idx)
		remove_edge_mask = torch.logical_or(temp_mask, remove_edge_mask)

	edge_index_aug = data.edge_index[:, ~remove_edge_mask]
	edge_num_aug = edge_index_aug.shape[1]
	edge_index_aug = edge_index_aug.numpy()
	edge_index_aug = [[idx_dict[edge_index_aug[0, n]], idx_dict[edge_index_aug[1, n]]] for n in range(edge_num_aug) if
	                  not edge_index_aug[0, n] == edge_index_aug[1, n]]
	try:
		edge_index_aug = torch.tensor(edge_index_aug).transpose_(0, 1)
	except:
		#         print(data)
		#         print(data.edge_index)
		#         print(idx_drop)
		#         print(idx_nondrop)
		#         print(idx_dict)
		#         print(~remove_edge_mask)
		#         print(edge_index_aug)
		#         print("Removal of all edges happened ... Returning original data")
		edge_index_aug = torch.tensor([[], []], dtype=torch.long)

	if hasattr(data, "edge_attr")  and data.edge_attr is not None:
		edge_attr_aug = data.edge_attr[~remove_edge_mask]
		data_aug = Data(x=x_aug, edge_index=edge_index_aug, edge_attr=edge_attr_aug)
	else:
		data_aug = Data(x=x_aug, edge_index=edge_index_aug)

	# edge_index = [[idx_dict[edge_index[0, n]], idx_dict[edge_index[1, n]]] for n in range(edge_num) if (not edge_index[0, n] in idx_drop) and (not edge_index[1, n] in idx_drop)]
	# edge_index = [[edge_index[0, n], edge_index[1, n]] for n in range(edge_num) if (not edge_index[0, n] in idx_drop) and (not edge_index[1, n] in idx_drop)] + [[n, n] for n in idx_nondrop]
	# data.edge_index = torch.tensor(edge_index).transpose_(0, 1)

	return data_aug


def perturb_edges(data, edge_ratio=0.1):
	node_num, _ = data.x.size()
	_, edge_num = data.edge_index.size()
	permute_num = int(edge_num * edge_ratio)

	edge_index = data.edge_index.transpose(0, 1).numpy()
	mask = np.random.choice(edge_num, edge_num - permute_num, replace=False)
	edge_index_aug = edge_index[mask]
	edge_index_aug = torch.tensor(edge_index_aug).transpose_(0, 1)

	if hasattr(data, "edge_attr") and data.edge_attr is not None:
		edge_attr_aug = data.edge_attr[mask]
		data_aug = Data(x=data.x, edge_index=edge_index_aug, edge_attr=edge_attr_aug)
	else:
		data_aug = Data(x=data.x, edge_index=edge_index_aug)

	return data_aug

def get_aug_pair(data):
	n = np.random.randint(2)
	if n == 0:
		data_aug = drop_nodes(data)
	elif n == 1:
		data_aug = subgraph(data)
	else:
		raise NotImplementedError
	return data, data_aug


def biased_perturb_edges(data, edge_ratio=0.1):
	node_num, _ = data.x.size()
	_, edge_num = data.edge_index.size()

	permute_num = int(edge_num * edge_ratio)
	core_edges_mask = data.core.numpy().astype(bool)

	edge_index = data.edge_index.transpose(0, 1).numpy()

	mask = np.random.choice(edge_num, edge_num - permute_num, replace=False)
	mask = np.union1d(mask, np.arange(edge_num)[core_edges_mask])
	edge_index_aug = edge_index[mask]
	edge_index_aug = torch.tensor(edge_index_aug).transpose_(0, 1)

	if hasattr(data, "edge_attr") and data.edge_attr is not None:
		edge_attr_aug = data.edge_attr[mask]
		data_aug = Data(x=data.x, edge_index=edge_index_aug, edge_attr=edge_attr_aug)
	else:
		data_aug = Data(x=data.x, edge_index=edge_index_aug)

	return data_aug

class BiasedAugTransformer():
	def __init__(self, type_code="pedges", aug_ratio=0.1):
		self.tc = type_code
		self.aug_ratio = aug_ratio
		# self.dn_ratio = aug_ratio
		# self.sub_ratio = aug_ratio
		# self.edge_ratio = aug_r atio
		print("Biased Augmentation Type: %s, Aug Ratio: %f, Using Edge Mask Name core " % (self.tc, self.aug_ratio))

	def __call__(self, data):
		if self.tc == "copy":
			data_aug = copy.deepcopy(data)
		elif self.tc == "pedges":
			data_aug = biased_perturb_edges(data, self.aug_ratio)
		else:
			raise NotImplementedError

		return data, data_aug


class MyAugTransformer(object):
	def __init__(self, type_code="dnodes", aug_ratio=0.1):
		self.tc = type_code
		self.aug_ratio = aug_ratio
		# self.dn_ratio = aug_ratio
		# self.sub_ratio = aug_ratio
		# self.edge_ratio = aug_ratio
		print("Augmentation Type: %s, Aug Ratio: %f " %(self.tc, self.aug_ratio))

	def __call__(self, data):
		if self.tc == "copy":
			data_aug = copy.deepcopy(data)
		elif self.tc == "dnodes":
			data_aug = drop_nodes(data, self.aug_ratio)
		elif self.tc == "pedges":
			data_aug = perturb_edges(data, self.aug_ratio)
		elif self.tc == "subgraph":
			data_aug = subgraph(data, self.aug_ratio)
		elif self.tc == "dnodes+subgraph":
			n = np.random.randint(2)
			if n == 0:
				data_aug = drop_nodes(data, self.aug_ratio)
			elif n == 1:
				data_aug = subgraph(data, self.aug_ratio)
			else:
				raise NotImplementedError
		else:
			raise NotImplementedError

		return data, data_aug


from typing import Callable, Union

import torch
from torch import Tensor
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.typing import OptPairTensor, Adj, Size

from unsupervised.convs.inits import reset


class WGINConv(MessagePassing):
	def __init__(self, nn: Callable, eps: float = 0., train_eps: bool = False,
				 **kwargs):
		kwargs.setdefault('aggr', 'add')
		super(WGINConv, self).__init__(**kwargs)
		self.nn = nn
		self.initial_eps = eps
		if train_eps:
			self.eps = torch.nn.Parameter(torch.Tensor([eps]))
		else:
			self.register_buffer('eps', torch.Tensor([eps]))
		self.reset_parameters()

	def reset_parameters(self):
		reset(self.nn)
		self.eps.data.fill_(self.initial_eps)

	def forward(self, x: Union[Tensor, OptPairTensor], edge_index: Adj, edge_weight = None,
				size: Size = None) -> Tensor:
		""""""
		if isinstance(x, Tensor):
			x: OptPairTensor = (x, x)

		# propagate_type: (x: OptPairTensor)
		out = self.propagate(edge_index, x=x, edge_weight=edge_weight, size=size)

		x_r = x[1]
		if x_r is not None:
			out += (1 + self.eps) * x_r

		return self.nn(out)

	def message(self, x_j: Tensor, edge_weight) -> Tensor:
		return x_j if edge_weight is None else x_j * edge_weight.view(-1, 1)


	def __repr__(self):
		return '{}(nn={})'.format(self.__class__.__name__, self.nn)


from .loaders import get_test_datasets, get_val_datasets, get_train_datasets, get_all_datasets
from .loaders import get_graph_task_datasets, get_edge_task_datasets, get_node_task_datasets
from .loaders import get_graph_classification_datasets, get_graph_regression_datasets, get_datasets


__all__ = ['get_test_datasets', 'get_val_datasets', 'get_train_datasets', 'get_datasets',
            'get_all_datasets', 'get_graph_task_datasets', 'get_edge_task_datasets',
              'get_node_task_datasets', 'get_graph_classification_datasets', 'get_graph_regression_datasets']


import torch
from torch.nn import Sequential, Linear, ReLU

class GSimCLR(torch.nn.Module):
	def __init__(self, encoder, proj_hidden_dim=300):
		super(GSimCLR, self).__init__()

		self.encoder = encoder
		self.input_proj_dim = self.encoder.out_graph_dim

		self.proj_head = Sequential(Linear(self.input_proj_dim, proj_hidden_dim), ReLU(inplace=True),
		                            Linear(proj_hidden_dim, proj_hidden_dim))

		self.init_emb()

	def init_emb(self):
		for m in self.modules():
			if isinstance(m, Linear):
				torch.nn.init.xavier_uniform_(m.weight.data)
				if m.bias is not None:
					m.bias.data.fill_(0.0)

	def forward(self, batch, x, edge_index, edge_attr=None, edge_weight=None):
		if edge_attr is not None:
			z, _ = self.encoder(batch, x, edge_index, edge_attr, edge_weight)
		else:
			z, _ = self.encoder(batch, x, edge_index, edge_weight)

		z = self.proj_head(z)
		# z shape -> Batch x (hidden_dim*num_gc_layers)
		return z

	@staticmethod
	def calc_loss(x, x_aug, temperature=0.2):
		# x shape -> Batch x (hidden_dim*num_gc_layers)

		batch_size, _ = x.size()
		x_abs = x.norm(dim=1)
		x_aug_abs = x_aug.norm(dim=1)

		sim_matrix = torch.einsum('ik,jk->ij', x, x_aug) / torch.einsum('i,j->ij', x_abs, x_aug_abs)
		sim_matrix = torch.exp(sim_matrix / temperature)
		pos_sim = sim_matrix[range(batch_size), range(batch_size)]
		loss = pos_sim / (sim_matrix.sum(dim=1) - pos_sim)
		loss = - torch.log(loss).mean()

		return loss


import argparse
import logging
import numpy as np
import torch
from torch_geometric.transforms import Compose
from bgd import get_train_datasets, get_val_datasets, get_test_datasets, get_all_datasets
from bgd.utils import *
from top import compute_top_scores, ToPDataset


def desc_datasets(datasets, stage, dataset_names):
    these_print_strings = []
    for i_dataset, dataset in enumerate(datasets):
        arrays, metrics, names = get_metric_values(dataset)
        print_string = f"{dataset_names[i_dataset]} & {stage} & {len(dataset)}"

        one_sample = dataset[0]

        x_shape, edge_shape, y_shape = one_sample.x, one_sample.edge_attr, one_sample.y
        x_shape = x_shape.shape[1] if x_shape is not None else "none" 
        edge_shape = edge_shape.shape[1] if edge_shape is not None else "none" 
        try:
            y_shape = y_shape.shape if y_shape is not None else "none" 
            if y_shape != "none":
                if len(y_shape) == 2:
                    y_shape = y_shape[1]
                else:
                    y_shape = y_shape[0]
        except:
            y_shape = "none"

        print_string += f" & {x_shape} & {edge_shape} & {y_shape} "
        for i_name, name in enumerate(names):
            value, dev = np.mean(arrays[i_name]), np.std(arrays[i_name])
            value = float('%.3g' % value)
            dev = float('%.3g' % dev)
            print_string += f"& {value} $\pm$ {dev}"
        print(print_string + r"\\")
        these_print_strings.append(print_string)

    print("\n")
    return these_print_strings

def latex_to_markdown(strings):
    new_string = ""
    for itxt, txt in enumerate(strings):
        
        txt = txt.replace("&", "|")
        txt = txt.replace(r"\\", "")
        txt = txt.replace("$", "")
        txt = txt.replace("\pm", "±")

        print("| " + txt + " |")
        new_string += "| " + txt + " | \n"

        if itxt == 0:
            print("|---|---|---|---|---|---|---|---|---|---|")
            new_string += "|---|---|---|---|---|---|---|---|---|---| \n"

    return new_string

def update_readme_table(readme_path, new_table):
    with open(readme_path, 'r') as file:
        readme = file.readlines()

    # Find the start and end of the table
    table_start = None
    table_end = None
    for i, line in enumerate(readme):
        if ' Name  ' in line:
            table_start = i - 1
        elif table_start is not None and line.strip() == '':
            table_end = i
            break

    if table_end is None:
        table_end = len(readme)

    if table_start is None or table_end is None:
        raise ValueError('Could not find table in README')

    # Split the new table into lines and add the newline character back to the end of each line
    new_table_lines = [line + '\n' for line in new_table.split('\n') if line]

    # Replace the table with the new table
    readme = readme[:table_start] + new_table_lines + readme[table_end:]
    
    # Write the updated README back to the file
    with open(readme_path, 'w') as file:
        file.write(''.join(readme))

def run(args):
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S')
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    logging.info("Using Device: %s" % device)
    logging.info("Seed: %d" % args.seed)
    logging.info(args)


    # Get datasets
    my_transforms = Compose([initialize_edge_weight])

    train_datasets, train_names = get_train_datasets(my_transforms, num = args.num_train)
    top_datasets = [ToPDataset(dataset) for i, dataset in enumerate(train_datasets)]
    compute_top_scores(top_datasets, train_names)

    roots = []
    for idataset, dataset in enumerate(train_datasets):
        try:
            roots.append(dataset.root)
        except AttributeError:
            train_datasets[idataset] = dataset.datasets[0]
            roots.append(train_datasets[idataset].root)
            # roots.append(dataset.datasets[0].root)

    for i, dataset in enumerate(train_datasets):
        print(dataset, train_names[i], roots[i])
        vis_grid(dataset[:25], f"{roots[i]}/{train_names[i]}.png")

        

    train_datasets, train_names = get_train_datasets(my_transforms, num = args.num_train)

    val_datasets, val_names = get_val_datasets(my_transforms, num = args.num_val)
    test_datasets, test_names = get_test_datasets(my_transforms, num = args.num_test)
    print_strings = [r"Name  &  Stage  &  Num  &  X shape  &  E shape  &  Y shape  &  Num. Nodes  &  Num. Edges  &  Diameter  &  Clustering \\"]
    print(print_strings[0])

    print_strings += desc_datasets(train_datasets, "Train", train_names)
    print_strings += desc_datasets(val_datasets, "Val", val_names)
    print_strings += desc_datasets(test_datasets, "Test", test_names)

    md_string = latex_to_markdown(print_strings)
    update_readme_table("README.md", md_string)








def arg_parse():
    parser = argparse.ArgumentParser(description='AD-GCL ogbg-mol*')

    parser.add_argument('--dataset', type=str, default='ogbg-molesol',
                        help='Dataset')

    parser.add_argument('--seed', type=int, default=0)

    parser.add_argument('--num_train', type=int, default=5000,
                        help='Number of points included in the train datasets')
    
    parser.add_argument('--num_val', type=int, default=1000,
                    help='Number of points included in the validation datasets')

    parser.add_argument('--num_test', type=int, default=1000,
                    help='Number of points included in the test datasets')
    
    parser.add_argument(
        '-s',
        '--score',
        action='store_true',
        help='Whether to compute similarity score against other datasets',
        default = False
    )

    return parser.parse_args()


if __name__ == '__main__':

    args = arg_parse()
    run(args)




import os
import networkx as nx
import torch
from tqdm import tqdm
from torch_geometric.data import InMemoryDataset, Data
from torch_geometric.utils.convert import to_networkx
from torch_geometric.io import read_npz
from torch.nn.functional import one_hot
import wget
import matplotlib.pyplot as plt
from ..utils import describe_one_dataset

if __name__ == "__main__":
    print(os.getcwd())

from ..utils import ESWR

def five_cycle_worker(g):
    """
    Returns the number of 5-cycles in a graph, normalised by the number of nodes
    """
    cycles = nx.simple_cycles(g, length_bound = 5)
    return len(list(cycles))

def vis_from_pyg(data, filename = None):
    edges = data.edge_index.T.cpu().numpy()
    labels = data.x[:,0].cpu().numpy()

    g = nx.Graph()
    g.add_edges_from(edges)

    fig, ax = plt.subplots(figsize = (6,6))

    pos = nx.kamada_kawai_layout(g)

    nx.draw_networkx_edges(g, pos = pos, ax = ax)
    nx.draw_networkx_nodes(g, pos = pos, node_color=labels, cmap="tab20",
                           vmin = 0, vmax = 20, ax = ax)

    ax.axis('off')

    plt.tight_layout()
    if filename is None:
        plt.show()
    else:
        plt.savefig(filename)
        plt.close()

def specific_from_networkx(graph):
    node_labels = []
    node_attrs = []
    edge_indices = []
    # Collect node labels and attributes
    for n in list(graph.nodes(data=True)):
        node_labels.append(n[1]["label"])
        node_attrs.append(n[1]["attrs"])

    # Collect edge indices and attributes
    for e in graph.edges(data=True):
        edge_indices.append((e[0], e[1]))

        # uncomment for edge attributes:
        # edge_attrs.append(e[2]["attr"]) 

    # Convert to PyTorch tensors
    node_labels = torch.stack(node_labels).flatten()

    # Specific to classification on nodes! Hard coding num classes as this happens on a per-graph basis
    node_labels = one_hot(node_labels.to(int), num_classes = 7)

    node_attrs = torch.stack(node_attrs)
    edge_indices = torch.tensor(edge_indices, dtype=torch.long).t().contiguous()

    # Create PyG Data object
    data = Data(x=node_attrs, edge_index=edge_indices, edge_attr = None,  y=node_labels)

    return data

def download_cora(visualise = False):
    # Ideally we'd save and reload graphs - but cora has massive feature dims
    zip_url = "https://github.com/abojchevski/graph2gauss/raw/master/data/cora_ml.npz"

    start_dir = os.getcwd()
    # print(os.getcwd(), os.listdir())
    os.chdir("bgd_files")

    if "cora" not in os.listdir():
        print("Downloading CORA")
        os.mkdir("cora")
        os.chdir("cora")
        _ = wget.download(zip_url)
    else:
        os.chdir("cora")
        if "cora_ml.npz" not in os.listdir():
            _ = wget.download(zip_url)

    edges = read_npz("cora_ml.npz")

    G = to_networkx(edges, to_undirected=True)

    node_classes = {n: int(edges.y[i].item()) for i, n in enumerate(list(G.nodes()))}
    node_attrs = {n: edges.x[i] for i, n in enumerate(list(G.nodes()))}
    # print(node_classes)

    # base_tensor = torch.Tensor([0])

    for node in list(G.nodes()):
        # class_tensor = base_tensor.clone()
        class_tensor = torch.Tensor([node_classes[node]])#node_classes[node]
        node_attr = node_attrs[node]

        # print(class_tensor, node_classes[node])

        G.nodes[node]["attrs"] = node_attr
        G.nodes[node]["label"] = class_tensor

    # for edge in list(G.edges()):
    #     G.edges[edge]["attrs"] = torch.Tensor([1])

    CGs = [G.subgraph(c) for c in nx.connected_components(G)]
    CGs = sorted(CGs, key=lambda x: x.number_of_nodes(), reverse=True)
    graph = CGs[0]
    graph = nx.convert_node_labels_to_integers(graph)

    os.chdir(start_dir)

    return graph





def get_cora_dataset(num = 2000, targets = False):
    cora_graph = download_cora()
    nx_graph_list = ESWR(cora_graph, num, 96)

    data_objects = [specific_from_networkx(graph) for graph in tqdm(nx_graph_list, desc = "Converting back to pyg graphs")]

    return  data_objects

class CoraDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`
    

    Academic citation graphs from the ML community, sampled from a large original graph using ESWR.
    The original graph is sourced from:

         `Yang, Zhilin, William Cohen, and Ruslan Salakhudinov. "Revisiting semi-supervised learning with graph embeddings." International conference on machine learning. PMLR, 2016.`

    The original data has one-hot bag-of-words over paper abstract as node features.

    The task is node classification for the category of each paper, one-hot encoded for seven categories.

     - Task: Node classification
     - Num node features: 2879
     - Num edge features: None
     - Num target values: 7
     - Target shape: N Nodes
     - Num graphs: Parameterised by `num`

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. (default: :obj:`2000`).
    """
    def __init__(self, root, stage = "train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        self.task = "node-classification"
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])


    @property
    def raw_file_names(self):
        return ['cora_ml.npz']

    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Cora files exist")
            return

        data_list = get_cora_dataset(num=self.num, targets=self.stage != "train")

        new_data_list = []
        for i, item in enumerate(data_list):
            n_nodes, n_edges = item.x.shape[0], item.edge_index.shape[1]


            data = Data(x = item.x,
                        edge_index=item.edge_index,
                        edge_attr=None,
                        y = item.y)

            new_data_list.append(data)
        data_list = new_data_list

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]


        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


if __name__ == "__main__":
    dataset = CoraDataset(os.getcwd()+'/bgd_files/'+'cora', stage = "train")
    describe_one_dataset(dataset)


    dataset = CoraDataset(os.getcwd()+'/bgd_files/'+'cora', stage = "val")
    describe_one_dataset(dataset)
    dataset = CoraDataset(os.getcwd()+'/bgd_files/'+'cora', stage = "test")
    describe_one_dataset(dataset)


import os
import networkx as nx
from networkx.algorithms.planarity import is_planar
import pandas as pd
import torch
from tqdm import tqdm
import torch_geometric as pyg
from torch_geometric.data import InMemoryDataset, Data
import pickle
import wget
import numpy as np
import gzip
import shutil
from ..utils import ESWR, describe_one_dataset
from littleballoffur.exploration_sampling import *

print(os.getcwd())

def four_cycles(g):
    """
    Returns the number of 4-cycles in a graph, normalised by the number of nodes
    """
    cycles = nx.simple_cycles(g, 4)
    return len(list(cycles)) / g.order()


def download_roads(visualise = False):
    zip_url = "https://snap.stanford.edu/data/roadNet-PA.txt.gz"

    start_dir = os.getcwd()
    os.chdir("bgd_files")


    if "roads" not in os.listdir():
        os.mkdir("roads")

    os.chdir("roads")
    # print(os.listdir())
    # quit()
    if "road_graph.npz" in os.listdir():
        with open("road_graph.npz", "rb") as f:
            graph = pickle.load(f)
        os.chdir(start_dir)
        return graph
    else:


        print("Downloading roads graph")
        # os.chdir("roads")

        _ = wget.download(zip_url)

        with gzip.open('roadNet-PA.txt.gz', 'rb') as f_in:
            with open('roadNet-PA.txt', 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        os.remove("roadNet-PA.txt.gz")
    # else:
    #     os.chdir("roads")

    #     if "road-graph.npz" in os.listdir():
    #         with open("road-graph.npz", "rb") as f:
    #             graph = pickle.load(f)
    #         os.chdir('../')
    #         return graph


    edgelist = pd.read_csv("roadNet-PA.txt", delimiter = "\t", header=3)

    graph = nx.Graph()

    sources = edgelist["# FromNodeId"].to_numpy().astype("int")
    targets = edgelist["ToNodeId"].to_numpy().astype("int")

    edges_to_add = [(sources[i], targets[i]) for i in tqdm(range(sources.shape[0]))]

    graph.add_edges_from(edges_to_add)

    del edges_to_add


    graph = nx.convert_node_labels_to_integers(graph)

    CGs = [graph.subgraph(c) for c in nx.connected_components(graph)]
    CGs = sorted(CGs, key=lambda x: x.number_of_nodes(), reverse=True)
    graph = CGs[0]
    graph = nx.convert_node_labels_to_integers(graph)
    graph.remove_edges_from(nx.selfloop_edges(graph))

    with open("road_graph.npz", "wb") as f:
        pickle.dump(graph, f)

    os.chdir(start_dir)
    return graph


def individual_sample(graph):
    sampler = DiffusionSampler(number_of_nodes=np.random.randint(12, 48))
    return nx.convert_node_labels_to_integers(sampler.sample(graph))

def add_attrs_to_graph(g):
    dummy_label = torch.Tensor([1])
    nx.set_edge_attributes(g, dummy_label, "attrs")
    nx.set_node_attributes(g, dummy_label, "attrs")


    return g


def get_road_dataset(num = 2000, targets = False):
    fb_graph = download_roads()
    nx_graph_list = ESWR(fb_graph, num, 96)


    data_objects = [pyg.utils.from_networkx(g) for g in nx_graph_list]

    for i, data in enumerate(tqdm(data_objects, desc="Calculating diameter values for roads", leave=False)):
        data.y = torch.tensor(int(is_planar(nx_graph_list[i])))

        nx_graph_list[i] = data

    return data_objects# loader

class RoadDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`
    

    Road graphs from Pennsylvania, sampled from a large original graph using ESWR.
    The original graph is sourced from:

         `J. Leskovec, K. Lang, A. Dasgupta, M. Mahoney. Community Structure in Large Networks: Natural Cluster Sizes and the Absence of Large Well-Defined Clusters. Internet Mathematics 6(1) 29--123, 2009.`

    The task is predicting whether a given graph is planar (can be laid out with no crossing edges).

     - Task: Graph classification
     - Num node features: None
     - Num edge features: None
     - Num target values: 1
     - Target shape: 1
     - Num graphs: Parameterised by `num`

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. (default: :obj:`2000`).
    """

    def __init__(self, root, stage="train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        _ = download_roads()
        del _
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return []


    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Road files exist")
            return

        data_list = get_road_dataset(num=self.num, targets=self.stage != "train")

        # if self.stage == "train":
        #     print("Found stage train, dropping targets")
        #     new_data_list = []
        #     for i, item in enumerate(data_list):
        #         n_nodes, n_edges = item.x.shape[0], item.edge_index.shape[1]

        #         data = Data(x = torch.ones(n_nodes).to(torch.int).reshape((-1, 1)),
        #                     edge_index=item.edge_index,
        #                     edge_attr=torch.ones(n_edges).to(torch.int).reshape((-1,1)),
        #                     y = None)
        #         new_data_list.append(data)
        #     data_list = new_data_list
        # else:
        #     new_data_list = []
        #     for i, item in enumerate(data_list):
        #         n_nodes, n_edges = item.x.shape[0], item.edge_index.shape[1]


        #         data = Data(x = item.x,
        #                     edge_index=item.edge_index,
        #                     edge_attr=torch.ones(n_edges).to(torch.int).reshape((-1,1)),
        #                     y = item.y)

        #         new_data_list.append(data)
        #     data_list = new_data_list

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])

        del data_list

if __name__ == "__main__":
    dataset = RoadDataset(os.getcwd()+'/bgd_files/'+'roads', stage = "train")
    describe_one_dataset(dataset)


import os
import torch
from torch_geometric.data import InMemoryDataset, Data
from ..utils import describe_one_dataset

from ogb.utils.features import get_atom_feature_dims, get_bond_feature_dims
from ogb.graphproppred import PygGraphPropPredDataset


full_atom_feature_dims = get_atom_feature_dims()
full_bond_feature_dims = get_bond_feature_dims()



def to_onehot_atoms(x):
    one_hot_tensors = []
    for i, num_values in enumerate(full_atom_feature_dims):
        one_hot = torch.nn.functional.one_hot(x[:, i], num_classes=num_values)
        one_hot_tensors.append(one_hot)

    return torch.cat(one_hot_tensors, dim=1)

def to_onehot_bonds(x):
    one_hot_tensors = []
    for i, num_values in enumerate(full_bond_feature_dims):
        one_hot = torch.nn.functional.one_hot(x[:, i], num_classes=num_values)
        one_hot_tensors.append(one_hot)

    return torch.cat(one_hot_tensors, dim=1)

class FromOGBDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`

    Converts an Open Graph Benchmark dataset into a `torch_geometric.data.InMemoryDataset`.
    This allows standard dataset operations like concatenation with other datasets.

    The Open Graph Benchmark project is available here:

         `Hu, Weihua, et al. "Open graph benchmark: Datasets for machine learning on graphs." Advances in neural information processing systems 33 (2020): 22118-22133.`

    We convert atom and bond features into one-hot encodings.
    The resulting shapes are:
     - node (atom features): (174, N Atoms)
     - edge (bond features) features: (13, N Bonds)

    Args:
        root (str): Root directory where the dataset should be saved.
        ogb_dataset (list): an `PygGraphPropPredDataset` to be converted back to `InMemoryDataset`.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. -1 takes all available samples for that stage. (default: :obj:`-1`).
    """
    def __init__(self, root, ogb_dataset, stage = "train", num = -1, transform=None, pre_transform=None, pre_filter=None):
        self.ogb_dataset = ogb_dataset
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2,
                               "train-adgcl":3}
        self.num = num
        print(f"Converting OGB stage {self.stage}")
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return ['dummy.csv']

    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.
        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print(f"\nTU files exist at {self.processed_paths[self.stage_to_index[self.stage]]}")
            return
        data_list = self.ogb_dataset

        num_samples = len(data_list)
        if num_samples < self.num:
            keep_n = num_samples
        else:
            keep_n = self.num

        new_data_list = []
        for i, item in enumerate(data_list[:keep_n]):


            data = Data(x = to_onehot_atoms(item.x), 
                        edge_index=item.edge_index,
                        edge_attr= to_onehot_bonds(item.edge_attr), 
                        y = item.y)
            
            new_data_list.append(data)
        data_list = new_data_list

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


def from_ogb_dataset(root,  stage="train", num=-1):
    """
    Load a dataset from the Open Graph Benchmark (OGB) and convert it to the Big Graph Dataset format.

    Args:
        name (str): The name of the OGB dataset. (Classification: "ogbg-molpcba", "ogbg-molhiv", "ogbg-moltox21", "ogbg-molbace", "ogbg-molbbbp", "ogbg-molclintox", "ogbg-molmuv", "ogbg-molsider", "ogbg-moltoxcast") (Regression: "ogbg-molesol", "ogbg-molfreesolv", "ogbg-mollipo")
        stage (str, optional): The stage of the dataset to load (e.g., "train", "valid", "test"). Defaults to "train".
        num (int, optional): The number of samples to load. Set to -1 to load all samples. Defaults to -1.

    Returns:
        FromOGBDataset: The converted dataset in the Big Graph Dataset format.
    """
    dataset = PygGraphPropPredDataset(root.split('/')[-1], root = root)
    split_idx = dataset.get_idx_split()
    dataset = dataset[split_idx["valid" if stage == "val" else stage]]
    return FromOGBDataset(root, dataset, stage = stage, num = num)

if __name__ == "__main__":
    molpcba = from_ogb_dataset("ogbg-molpcba", stage = "train")
    describe_one_dataset(molpcba)


import numpy as np
import networkx as nx
import torch
import torch_geometric as pyg
from torch_geometric.utils.convert import to_networkx
from tqdm import tqdm
import os
from torch_geometric.data import InMemoryDataset

def get_community_graph(size = 48, proportions = [0.25, 0.25, 0.25, 0.25], P_intra = 0.5, P_inter=0.05 + 0.1*np.random.random()):

    sizes = (np.array(proportions) * size).astype(int).tolist()#


    subgraphs = []
    counter = 0
    for i_size, size in enumerate(sizes):
        g = nx.Graph()

        for i in range(counter, counter + size):
            g.add_node(i)

        counter += size
        subgraphs.append(g)

    for g in subgraphs:
        for n1 in g.nodes():
            for n2 in g.nodes():
                if np.random.random() <= P_intra:
                    g.add_edge(n1, n2)

    node_identifiers = [list(g.nodes()) for g in subgraphs]

    G = nx.Graph()
    for g in subgraphs:
        G = nx.compose(G, g)

    for ids_1 in node_identifiers:
        for ids_2 in node_identifiers:
            if ids_1 == ids_2:
                pass
            else:
                for n1 in ids_1:
                    for n2 in ids_2:
                        if np.random.random() <= P_inter:
                            G.add_edge(n1, n2)

    return G, P_inter


def get_community_dataset(num = 1000):
    nx_graph_list_rhos = [get_community_graph(P_inter=0.05 + 0.05*np.random.random()) for _ in tqdm(range(num), leave=False)]
    nx_graph_list = [item[0] for item in nx_graph_list_rhos]
    rhos= [item[1] for item in nx_graph_list_rhos]

    datalist = [pyg.utils.from_networkx(g) for g in tqdm(nx_graph_list)]

    for idata, data in enumerate(datalist):
        data.y = torch.Tensor([rhos[idata]])
        datalist[idata] = data

    return datalist

class CommunityDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`

    Dataset of random community-like graphs, composed of four evenly sized subgraphs with a random inter-subgraph edge probability.
    Subgraphs have a density of 0.5, and inter-subgraph edges have a random density between 0.05 and 0.15.

    The target is the inter-subgraph density of each graph.

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. -1 takes all available samples for that stage. (default: :obj:`-1`).
    """
    def __init__(self, root, stage="train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        # _ = download_facebook()
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return []


    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Facebook files exist")
            return

        data_list = get_community_dataset(num=self.num)

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]


        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


import os
import pickle
import wget
import pandas as pd
import networkx as nx
from tqdm import tqdm
import numpy as np
from ..utils import ESWR
import torch
from torch_geometric.data import Data, InMemoryDataset


def fix_property_string(input_string):
    input_string = input_string.split(',')
    input_string = [float(item) for item in input_string]

    return np.array(input_string)

def download_reddit():
    print("Getting reddit networkx graph")
    start_dir = os.getcwd()
    print(start_dir)
    os.chdir("reddit")
    print(os.getcwd())

    graph_url = "https://snap.stanford.edu/data/soc-redditHyperlinks-title.tsv"
    embedding_url = "http://snap.stanford.edu/data/web-redditEmbeddings-subreddits.csv"
        
    if "reddit-graph.npz" in os.listdir():
        with open("reddit-graph.npz", "rb") as f:
            graph = pickle.load(f)
        os.chdir(start_dir)
        return graph

    if "soc-redditHyperlinks-title.tsv" not in os.listdir():
        graph_data = wget.download(graph_url)
    if "web-redditEmbeddings-subreddits.csv" not in os.listdir():
        embeddings = wget.download(embedding_url)

    # We know that there are 300 components in the node feature vectors
    embedding_column_names = ["COMPONENT", *[i for i in range(300)]]
    embeddings = pd.read_csv("web-redditEmbeddings-subreddits.csv", names=embedding_column_names).transpose()
    graph_data = pd.read_csv("soc-redditHyperlinks-title.tsv", sep = "\t")




    embeddings.columns = embeddings.iloc[0]
    embeddings = embeddings.drop(["COMPONENT"], axis = 0)

    graph = nx.Graph()

    for col in tqdm(embeddings.columns, desc = "Adding nodes"):
        # attrs here is taken from the embedding data, with the node id the column (col)
        graph.add_node(col, attrs=embeddings[col].to_numpy().astype(float))

    sources = graph_data["SOURCE_SUBREDDIT"].to_numpy()
    targets = graph_data["TARGET_SUBREDDIT"].to_numpy()

    # This line can take a while!
    attrs = [fix_property_string(properties) for properties in tqdm(graph_data["PROPERTIES"].tolist(), desc = "Wrangling edge features")]
    labels = graph_data["LINK_SENTIMENT"].to_numpy()
    all_nodes = list(graph.nodes())

    for i in tqdm(range(sources.shape[0]), desc = "Adding edges"):
        if sources[i] in all_nodes and targets[i] in all_nodes:
            graph.add_edge(sources[i], targets[i],
                        labels = labels[i],
                        attr = attrs[i])


    # Last tidying bits
    graph = nx.convert_node_labels_to_integers(graph)
    CGs = [graph.subgraph(c) for c in nx.connected_components(graph)]
    CGs = sorted(CGs, key=lambda x: x.number_of_nodes(), reverse=True)
    graph = CGs[0]
    graph = nx.convert_node_labels_to_integers(graph)

    # Save the graph!
    with open("bgd_files/reddit/reddit-graph.npz", "wb") as f:
        pickle.dump(graph, f)

        # Avoids weird directory problems
    os.chdir(start_dir)

    return graph

def get_reddit_dataset(num = 2000):
    graph = download_reddit()

    # Sample 1000 graphs of max 96 nodes from the big reddit graph
    nx_graph_list = ESWR(graph, 1000, 96)

    pyg_graph_list = [specific_from_networkx(g) for g in nx_graph_list]

    return pyg_graph_list

def specific_from_networkx(graph):
    # Turns a graph into a pytorch geometric object
    # Mostly by unpacking dictionaries on nodes and edges
    # Here node labels are the target
    # One of these functions for each dataset ideally - they are unlikely to transfer across datasets
    node_attrs = []
    edge_indices = []
    edge_labels = []
    edge_attrs = []

    # Collect node labels and attributes
    for n in list(graph.nodes(data=True)):
        # list(graph.nodes(data=True)) returns [(node_id1, {attribute dictionary}), (node_id2, ...), (node_id3, ...)]
        node_attrs.append(torch.Tensor(n[1]["attrs"]))

    # Collect edge indices and attributes
    for e in graph.edges(data=True):
        # graph.edges(data=True) is a generator producing (node_id1, node_id2, {attribute dictionary})
        edge_indices.append((e[0], e[1]))

        edge_attrs.append(torch.Tensor(e[2]["attr"])) 
        edge_labels.append(e[2]["labels"])


    # Specific to classification on edges! This is a binary edge classification (pos/neg) task
    edge_labels = ((torch.Tensor(edge_labels) + 1)/2).reshape(-1,1)

    edge_attrs = torch.stack(edge_attrs)
    node_attrs = torch.stack(node_attrs)
    edge_indices = torch.tensor(edge_indices, dtype=torch.long).t().contiguous()

    # Create PyG Data object
    # Can pass:
    # x:            node features, shape (n nodes x n features)
    # edge_index:   the list of edges in the graph, shape (2, n_edges). Entries edge_index[i, :] are [node_id1, node_id2].
    # edge_attr:    edge features, shape (n_edges, n_features), same order as edgelist
    # y:            targets. Graph regression shape (n_variables), graph classification (n_classes), node classification (n_nodes, n_classes), edge classification (n_edges, n_classes)
    data = Data(x=node_attrs, edge_index=edge_indices, edge_attr = edge_attrs,  y=edge_labels)

    return data


class RedditDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`
    

    Reddit hyperlink graphs - ie graphs of subreddits interacting with one another.
    The original graph is sourced from:

         `Kumar, Srijan, et al. "Community interaction and conflict on the web." Proceedings of the 2018 world wide web conference. 2018.`

    We produce this dataset of small graphs using ESWR.
    The data has text embeddings as node features for each subreddit and text features for the cross-post edges.

    The task is edge classification for the sentiment of the interaction between subreddits.

     - Task: Edge classification
     - Num node features: 300
     - Num edge features: 86
     - Num target values: 1
     - Target shape: N Edges
     - Num graphs: Parameterised by `num`

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. (default: :obj:`2000`).
    """

    def __init__(self, root,  stage = "train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        
        # Options are node-classification, node-regression, graph-classification, graph-regression, edge-regression, edge-classification
        # Graph-level tasks are preferred! (graph-classification and graph-regression)
        # edge-prediction is another option if you can't think of a good task
        self.task = "edge-classification"

        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])


    @property
    def raw_file_names(self):
        # Replace with your saved raw file name
        return ['reddit-graph.npz']

    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print(f"Reddit files exist")
            return

        # Get a list of num pytorch_geometric.data.Data objects
        data_list = get_reddit_dataset(self.num)

        # Torch geometric stuff
        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        # Save data
        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


import numpy as np
import networkx as nx
import torch
import os
import torch_geometric as pyg
from torch_geometric.utils.convert import to_networkx
from tqdm import tqdm
import matplotlib.pyplot as plt
from torch_geometric.data import Data, InMemoryDataset



def get_tree_graph(max_nodes = 96):



    # base_tensor = torch.Tensor([np.random.randint(1,9)])

    n = np.random.randint(low=8, high = max_nodes)
    G = nx.random_tree(n = n)
    # G_attr = nx.Graph()

    # for i in range(G.order()):
    #     G_attr.add_node(i)

    # for (n1, n2) in G.edges():
    # # for n1 in G.nodes():
    # #     for n2 in G.nodes():
    #     G_attr.add_edge(n1, n2)

    depth = nx.eccentricity(G, 0) / G.order()
    return G, depth


def get_tree_dataset(num = 1000):
    nx_graph_list_rhos = [get_tree_graph() for _ in tqdm(range(num), leave=False)]
    nx_graph_list = [item[0] for item in nx_graph_list_rhos]
    depths= [item[1] for item in nx_graph_list_rhos]
    Ns = [graph.order() for graph in nx_graph_list]

    datalist = [pyg.utils.from_networkx(g) for g in tqdm(nx_graph_list)]

    for idata, data in enumerate(datalist):
        data.y = torch.Tensor([depths[idata]])
        datalist[idata] = data

    return datalist


class TreeDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`

    Dataset of random tree structures, between 8 and 96 nodes, produced with `networkx.random_tree`.

    The target is the depth of the tree, normalised by the number of nodes in the tree.

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. -1 takes all available samples for that stage. (default: :obj:`-1`).
    """
    def __init__(self, root, stage="train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        # _ = download_facebook()
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return []


    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Tree files exist")
            return

        data_list = get_tree_dataset(num=self.num)

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


import os
import networkx as nx
import pandas as pd
import torch
from torch.nn.functional import one_hot
from torch_geometric.data import InMemoryDataset, Data
import pickle
import zipfile
import wget
import numpy as np
from ..utils import ESWR
import json

from littleballoffur.exploration_sampling import *

from ..utils import describe_one_dataset, vis_grid


def download_facebook(visualise = False):
    zip_url = "https://snap.stanford.edu/data/facebook_large.zip"

    start_dir = os.getcwd()
    os.chdir("bgd_files")

    if "facebook-graph.npz" in os.listdir():
        with open("facebook-graph.npz", "rb") as f:
            graph = pickle.load(f)
        os.chdir('../')
        return graph

    if "musae_facebook_edges.csv" not in os.listdir("facebook_large"):
        print("Downloading FB graph")
        _ = wget.download(zip_url)
        with zipfile.ZipFile("facebook_large.zip", 'r') as zip_ref:
            zip_ref.extractall(".")
        os.remove("facebook_large.zip")
    os.chdir("facebook_large")

    edgelist = pd.read_csv("musae_facebook_edges.csv")
    labels = pd.read_csv("musae_facebook_target.csv")

    # Not using node features as they are all different lengths :(
    with open("musae_facebook_features.json", 'r', encoding='utf-8') as file:
            features = json.load(file)


    conversion_dict = {"company":       torch.Tensor([0]), #, 0, 0, 0, 0, 0, 0, 0, 0]),
                       "government":    torch.Tensor([1]), #, 0, 0, 0, 0, 0, 0, 0, 0]),
                       "politician":    torch.Tensor([2]), #, 0, 0, 0, 0, 0, 0, 0, 0]),
                       "tvshow":        torch.Tensor([3])} #, 0, 0, 0, 0, 0, 0, 0, 0]),


    graph = nx.Graph()
    label_specific = labels["page_type"]
    for col in labels["id"]:
        graph.add_node(int(col))
        # Turns out the facebook data has attributes of varying length
        graph.nodes[int(col)]["label"] = conversion_dict[label_specific[col]]
        # graph.nodes[int(col)]["attrs"] = torch.Tensor(features[str(col)])

    sources = edgelist["id_1"].to_numpy().astype("int")
    targets = edgelist["id_2"].to_numpy().astype("int")

    for i in range(sources.shape[0]):
        graph.add_edge(sources[i], targets[i])

    for node in list(graph.nodes(data=True)):
        data = node[1]
        if len(data) == 0:
            graph.remove_node(node[0])

    graph = nx.convert_node_labels_to_integers(graph)

    CGs = [graph.subgraph(c) for c in nx.connected_components(graph)]
    CGs = sorted(CGs, key=lambda x: x.number_of_nodes(), reverse=True)
    graph = CGs[0]
    graph = nx.convert_node_labels_to_integers(graph)
    graph.remove_edges_from(nx.selfloop_edges(graph))

    os.chdir(start_dir)
    return graph

def specific_from_networkx(graph):
    node_labels = []
    edge_indices = []
    # node_attrs = []

    # Collect node labels and attributes
    for n in list(graph.nodes(data=True)):
        node_labels.append(n[1]["label"])
        # Weird features, different sizes, so not including
        # node_attrs.append(n[1]["attrs"])

    # Collect edge indices and attributes
    for e in graph.edges(data=True):
        edge_indices.append((e[0], e[1]))

        # uncomment for edge attributes:
        # edge_attrs.append(e[2]["attr"]) 

    # Convert to PyTorch tensors
    node_labels = torch.stack(node_labels).flatten()

    # Specific to classification on nodes! Hard coding num classes as this happens on a per-graph basis
    node_labels = one_hot(node_labels.to(int), num_classes = 4)

    # node_attrs = torch.stack(node_attrs)
    edge_indices = torch.tensor(edge_indices, dtype=torch.long).t().contiguous()

    # Create PyG Data object
    data = Data(x=None, edge_index=edge_indices, edge_attr = None,  y=node_labels)

    return data

def get_fb_dataset(num = 2000, targets = False):
    fb_graph = download_facebook()
    # print(fb_graph.nodes(data=True))
    nx_graph_list = ESWR(fb_graph, num, 96)
    data_objects = [specific_from_networkx(g) for g in nx_graph_list]

    return data_objects# loader

class FacebookDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`
    

    Facebook page-to-page interaction graphs, sampled from a large original graph using ESWR.
    The original graph is sourced from:

         `Benedek Rozemberczki, Carl Allen, and Rik Sarkar. Multi-Scale Attributed Node Embedding.  Journal of Complex Networks 2021`

    The original data has node features, but as they are of varying length, we don't include them here.

    The task is node classification for the category of each Facebook page in a given graph, one-hot encoded for four categories.

     - Task: Node classification
     - Num node features: None
     - Num edge features: None
     - Num target values: 4
     - Target shape: N Nodes
     - Num graphs: Parameterised by `num`

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. (default: :obj:`2000`).
    """
        
    def __init__(self, root, stage="train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        
        self.task = "node-classification"

        # _ = download_facebook()
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return ['musae_facebook_edges.csv',
                'musae_facebook_target.csv']


    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Facebook files exist")
            return

        data_list = get_fb_dataset(num=self.num, targets=self.stage != "train")

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]


        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


if __name__ == "__main__":
    dataset = FacebookDataset(os.getcwd()+'/bgd_files/'+'facebook_large', stage = "train", num = 5000)
    describe_one_dataset(dataset)
    vis_grid(dataset[:16], os.getcwd()+"/bgd_files/facebook_large/train.png")
    
    dataset = FacebookDataset(os.getcwd()+'/bgd_files/'+'facebook_large', stage = "val", num = 1000)
    describe_one_dataset(dataset)
    vis_grid(dataset[:16], os.getcwd()+"/bgd_files/facebook_large/val.png")
    
    dataset = FacebookDataset(os.getcwd()+'/bgd_files/'+'facebook_large', stage = "test", num = 1000)
    describe_one_dataset(dataset)
    vis_grid(dataset[:16], os.getcwd()+"/bgd_files/facebook_large/test.png")


from .cora_dataset import CoraDataset
from .road_dataset import RoadDataset
from .neural_dataset import NeuralDataset
from .facebook_dataset import FacebookDataset
from .reddit_dataset import RedditDataset
from .ego_dataset import EgoDataset
from .from_ogb_dataset import FromOGBDataset, from_ogb_dataset
from .from_tu_dataset import FromTUDataset, from_tu_dataset

__all__ = ['CoraDataset', 'RoadDataset', 'NeuralDataset',
            'FacebookDataset', 'RedditDataset', 'EgoDataset',
            'FromOGBDataset', 'from_ogb_dataset',
            'FromTUDataset', 'from_tu_dataset']


import numpy as np
import networkx as nx
import torch
import os
import torch_geometric as pyg
from torch_geometric.utils.convert import to_networkx
from tqdm import tqdm
import matplotlib.pyplot as plt
from torch_geometric.data import Data, InMemoryDataset

# Extra keywords for compatibility
def line_graph(n = 12, m = 12, create_using="dummy"):

    G = nx.Graph()
    N = n * m
    for i_n in range(N - 1):
        G.add_edge(i_n - 1, i_n)

    return G

def get_lattice_graph(max_nodes = 16):



    # base_tensor = torch.Tensor([np.random.randint(1,9)])

    n = np.random.randint(low=3, high = max_nodes)
    m = np.random.randint(low=3, high=max_nodes)


    generators = [line_graph, nx.triangular_lattice_graph, nx.grid_2d_graph, nx.hexagonal_lattice_graph]
    lattice_selection = np.random.randint(low=0, high=4)
    lattice_generator = generators[lattice_selection]

    G = nx.convert_node_labels_to_integers(lattice_generator(m = m, n = n, create_using=nx.Graph))
    G_attr = nx.Graph()

    for i in range(G.order()):
        G_attr.add_node(i, attr = torch.Tensor([1]))

    for (n1, n2) in G.edges():
    # for n1 in G.nodes():
    #     for n2 in G.nodes():
        G_attr.add_edge(n1, n2, attr=torch.Tensor([1]))

    return G_attr, lattice_selection


def get_lattice_dataset(keep_target = False, num = 1000):
    nx_graph_list_rhos = [get_lattice_graph() for _ in tqdm(range(num), leave=False)]
    nx_graph_list = [item[0] for item in nx_graph_list_rhos]
    depths= [item[1] for item in nx_graph_list_rhos]
    Ns = [graph.order() for graph in nx_graph_list]

    datalist = [pyg.utils.from_networkx(g, group_node_attrs=all, group_edge_attrs=all) for g in tqdm(nx_graph_list)]

    if keep_target:
        for idata, data in enumerate(datalist):
            data.y = torch.Tensor([depths[idata]])
            datalist[idata] = data

    return datalist


class LatticeDataset(InMemoryDataset):
    def __init__(self, root, stage="train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}
        # _ = download_facebook()
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return []


    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print("Tree files exist")
            return

        data_list = get_lattice_dataset(num=self.num, keep_target=self.stage != "train")#get_fb_dataset(num=self.num)

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        # if self.stage != "train":
        #     for i, data in enumerate(data_list):
        #         vis_from_pyg(data, filename=self.root + f'/processed/{self.stage}-{i}.png')

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


import os
from torch_geometric.data import DataLoader

from ogb.graphproppred import PygGraphPropPredDataset
from bgd.real.facebook_dataset import FacebookDataset
from bgd.real.ego_dataset import EgoDataset
from bgd.real.cora_dataset import CoraDataset
from bgd.real.neural_dataset import NeuralDataset
from bgd.real.road_dataset import RoadDataset
from bgd.real.reddit_dataset import RedditDataset
from bgd.real.from_ogb_dataset import FromOGBDataset, from_ogb_dataset
from bgd.real.from_tu_dataset import from_tu_dataset

from bgd.synthetic.random_dataset import RandomDataset
from bgd.synthetic.tree_dataset import TreeDataset
from bgd.synthetic.lattice_dataset import LatticeDataset
from bgd.synthetic.community_dataset import CommunityDataset


def get_datasets(transforms, num, stage="train", exclude=None, include=None):
    r"""
    Retrieves and transforms a list of datasets based on specified inclusion and exclusion criteria.

    Parameters:
        transforms (function): A function to apply transformations to each dataset.
        num (int): The number of data points to include in each dataset.
        stage (str, optional): The stage of data processing (e.g., "train", "test", "validate"). Default is "train".
        exclude (list or str, optional): A list or a single string specifying dataset names to exclude from the selection. If None, no datasets will be excluded. Default is None.
        include (list or str, optional): A list or a single string specifying dataset names to include in the selection. If None, all datasets not in the exclude list will be included. Default is None.

    Returns:
        list: A list of transformed datasets.
        list: A list of names of the selected datasets.

    Notes:
        - If both `exclude` and `include` are provided, the function first applies the `exclude` filter and then the `include` filter.
        - The function checks for the existence of a "bgd_files" directory and creates it if it does not exist.
        - The function supports various datasets, including predefined datasets and those from the Open Graph Benchmark (OGB) and TU datasets.

    Example:
        >>> def dummy_transform(dataset):
        >>>     return dataset
        >>> datasets, names = get_datasets(dummy_transform, num=100, stage="train", exclude=["reddit"], include=["cora", "trees"])
        >>> print(names)
        ['cora', 'trees']
    """


    if "bgd_files" not in os.listdir():
        os.mkdir("bgd_files")
        
    all_datasets = {
        "facebook_large": FacebookDataset,
        "twitch_egos": EgoDataset,
        "cora": CoraDataset,
        "roads": RoadDataset,
        "fruit_fly": NeuralDataset,
        "reddit": RedditDataset,
        "trees": TreeDataset,
        "random": RandomDataset,
        "community": CommunityDataset
    }

    ogb_names = ["ogbg-molpcba", "ogbg-molesol", "ogbg-molclintox",
                 "ogbg-molfreesolv", "ogbg-mollipo", "ogbg-molhiv",
                 "ogbg-molbbbp", "ogbg-molbace"]
    all_datasets.update({name: from_ogb_dataset for name in ogb_names})

    tu_names = ["MUTAG", "ENZYMES", "PROTEINS", "COLLAB", "IMDB-BINARY", "REDDIT-BINARY"]
    all_datasets.update({name: from_tu_dataset for name in tu_names})

    names = list(all_datasets.keys())

    if exclude:
        if isinstance(exclude, str):
            exclude = [exclude]
        names = [d for d in names if d not in exclude]

    if include:
        if isinstance(include, str):
            include = [include]
        names = [d for d in names if d in include]

    datasets = [transforms(all_datasets[name](os.path.join(os.getcwd(), 'bgd_files', name), num=num, stage=stage))
                for name in names]

    return datasets, names

def get_node_task_datasets(transforms, num=5000, stage="train"):
    """
    Returns datasets with node-level tasks, both classification and regression.

    Args:
        transforms (list): List of data transformations to apply to the datasets.
        num (int, optional): Number of datasets to retrieve. Defaults to 5000.
        stage (str, optional): Stage of the datasets to retrieve. Defaults to "train".

    Returns:
        list: List of node task datasets (:obj:torch_geometric.data.InMemoryDataset).

    """
    includes = ["facebook_large", "cora"]
    return get_datasets(transforms, num, stage, include=includes)

def get_edge_task_datasets(transforms, num=5000, stage="train"):
    """
    Returns datasets with edge-level tasks, both regression and classification.

    Args:
        transforms (list): List of data transformations to apply to the datasets.
        num (int, optional): Number of datasets to retrieve. Defaults to 5000.
        stage (str, optional): Stage of the datasets to retrieve. Defaults to "train".

    Returns:
        list: List of edge task datasets (:obj:torch_geometric.data.InMemoryDataset).

    """
    includes = ["fruit_fly", "reddit"]
    return get_datasets(transforms, num, stage, include=includes)

def get_graph_task_datasets(transforms, num=5000, stage="train"):
    """
    Returns datasets with graph-level tasks, both regression and classification.

    Args:
        transforms (list): List of data transformations to apply to the datasets.
        num (int, optional): Number of datasets to retrieve. Defaults to 5000.
        stage (str, optional): Stage of the datasets to retrieve. Defaults to "train".

    Returns:
        list: List of graph task datasets (:obj:torch_geometric.data.InMemoryDataset).

    """
    non_graph_level_excludes = ["facebook_large", "cora", "fruit_fly", "reddit"]
    return get_datasets(transforms, num, stage, exclude=non_graph_level_excludes)

def get_graph_classification_datasets(transforms, num=5000, stage="train"):
    """
    Returns datasets with graph classification tasks.

    Args:
        transforms (list): List of data transformations to apply to the datasets.
        num (int, optional): Number of datasets to retrieve. Defaults to 5000.
        stage (str, optional): Stage of the datasets to retrieve. Defaults to "train".

    Returns:
        list: List of graph classification datasets (:obj:torch_geometric.data.InMemoryDataset).

    """
    non_graph_classification_excludes = ["facebook_large", "cora", "fruit_fly", "reddit",
                                         "ogbg-molesol", "ogbg-molfreesolv", "ogbg-mollipo",
                                         "community", "trees", "random"]
    return get_datasets(transforms, num, stage, exclude=non_graph_classification_excludes)

def get_graph_regression_datasets(transforms, num=5000, stage="train"):
    """
    Returns datasets with graph regression tasks.

    Args:
        transforms (list): List of data transformations to apply to the datasets.
        num (int, optional): Number of datasets to retrieve. Defaults to 5000.
        stage (str, optional): Stage of the datasets to retrieve. Defaults to "train".

    Returns:
        list: List of graph regression datasets (:obj:torch_geometric.data.InMemoryDataset).

    """
    non_graph_level_excludes = ["facebook_large", "cora", "fruit_fly", "reddit",
                                "ogbg-molpcba", "ogbg-molhiv", "ogbg-moltox21",
                                "ogbg-molbace", "ogbg-molbbbp", "ogbg-molclintox",
                                "ogbg-molmuv", "ogbg-molsider", "ogbg-moltoxcast",
                                "MUTAG", "ENZYMES", "PROTEINS",
                                "COLLAB", "IMDB-BINARY", "REDDIT-BINARY"]
    return get_datasets(transforms, num, stage, exclude=non_graph_level_excludes)

def get_test_datasets(transforms, num=2000, mol_only=False, exclude=["community", "trees", "random"], include=None):
    """
    Get the test split of each dataset.

    Args:
        transforms (list): List of data transformations to apply.
        num (int): Number of samples in datasets to include (default is 2000).
        mol_only (bool): Flag indicating whether to include only chemical datasets (default is False).

    Returns:
        tuple: A tuple containing two elements:
            - datasets (list): List of test datasets.
            - names (list): List of dataset names.
    """

    datasets, names = get_datasets(transforms, num, stage="test", include=include, exclude=exclude)
    return datasets, names


def get_val_datasets(transforms, num=2000, mol_only=False, exclude=["community", "trees", "random"], include=None):
    """
    Get validation splits for each dataset.

    Args:
        transforms (list): List of data transformations to apply.
        num (int, optional): Number of samples in datasets to include. Defaults to 2000.
        mol_only (bool, optional): Flag indicating whether to include only chemical datasets. Defaults to False.

    Returns:
        tuple: A tuple containing two elements:
            - datasets (list): List of validation datasets.
            - names (list): List of dataset names.
    """

    datasets, names = get_datasets(transforms, num, stage="val", include=include, exclude=exclude)
    return datasets, names

def get_train_datasets(transforms, num=2000, mol_only=False, exclude=["ogbg-molpcba"], include=None):
    """
    Get the training splits of each dataset.

    Args:
        transforms (list): List of data transformations to apply.
        num (int): Number of datasets to retrieve.
        mol_only (bool): Flag indicating whether to retrieve only chemical datasets.

    Returns:
        tuple: A tuple containing two elements:
            - datasets (list): A list of all the datasets.
            - all_names (list): A list of names corresponding to each dataset.
    """

    datasets, names = get_datasets(transforms, num, stage= "train", exclude = exclude, include = include)
    return datasets, names

def remove_duplicates_in_place(list1, list2):
    seen = set()
    i = 0
    
    while i < len(list2):
        if list2[i] in seen:
            del list1[i]
            del list2[i]
        else:
            seen.add(list2[i])
            i += 1

def get_all_datasets(transforms, num=5000, mol_only=False):
    """
    Get all datasets for training and validation, in that order.

    Args:
        transforms (list): List of data transformations to apply to the datasets.
        num (int, optional): Number of samples to load from each dataset. Defaults to 5000.
        mol_only (bool, optional): Flag indicating whether to include only chemical datasets. Defaults to False.

    Returns:
        tuple: A tuple containing two elements:
            - datasets (list): A list of all the datasets.
            - all_names (list): A list of names corresponding to each dataset.
    """

    train_datasets, train_names = get_train_datasets(transforms, num)
    val_datasets, val_names = get_val_datasets(transforms, -1)
    test_datasets, test_names = get_val_datasets(transforms, -1)

    datasets = train_datasets + val_datasets + test_datasets
    all_names = train_names + val_names + test_names

    remove_duplicates_in_place(datasets, all_names)

    return datasets, all_names




import numpy as np
import torch
import torch.nn.functional as F
from ogb.graphproppred.mol_encoder import AtomEncoder, BondEncoder
from torch.nn import Sequential, Linear, ReLU
from torch_geometric.nn import global_add_pool
from unsupervised.convs import GINEConv
from bgd.utils import normalize_l2


class Encoder(torch.nn.Module):
	"""
	Encoder module for graph neural networks.

	Args:
		emb_dim (int): The dimensionality of the node embeddings.
		num_gc_layers (int): The number of graph convolutional layers.
		drop_ratio (float): The dropout ratio.
		pooling_type (str): The type of graph pooling to use.
		is_infograph (bool): Whether to use Infograph pooling.
		convolution (str): The type of graph convolutional operation to use.
		edge_dim (int): The dimensionality of the edge embeddings.

	Attributes:
		pooling_type (str): The type of graph pooling being used.
		emb_dim (int): The dimensionality of the node embeddings.
		num_gc_layers (int): The number of graph convolutional layers.
		drop_ratio (float): The dropout ratio.
		is_infograph (bool): Whether to use Infograph pooling.
		out_node_dim (int): The output dimensionality of the node embeddings.
		out_graph_dim (int): The output dimensionality of the graph embeddings.
		convs (torch.nn.ModuleList): List of graph convolutional layers.
		bns (torch.nn.ModuleList): List of batch normalization layers.
		atom_encoder (AtomEncoder): Atom encoder module.
		bond_encoder (BondEncoder): Bond encoder module.
		edge_dim (int): The dimensionality of the edge embeddings.
		convolution (type): The type of graph convolutional operation being used.

	Methods:
		init_emb(): Initializes the node embeddings.
		forward(batch, x, edge_index, edge_attr, edge_weight=None): Performs forward pass through the encoder.
		get_embeddings(loader, device, is_rand_label=False, every=1, node_features=False): Computes embeddings for a given data loader.

	"""

	def __init__(self, emb_dim=300, num_gc_layers=5, drop_ratio=0.0,
				 pooling_type="standard", is_infograph=False,
				 convolution="gin", edge_dim=1):
		super(Encoder, self).__init__()

		self.pooling_type = pooling_type
		self.emb_dim = emb_dim
		self.num_gc_layers = num_gc_layers
		self.drop_ratio = drop_ratio
		self.is_infograph = is_infograph

		self.out_node_dim = self.emb_dim
		if self.pooling_type == "standard":
			self.out_graph_dim = self.emb_dim
		elif self.pooling_type == "layerwise":
			self.out_graph_dim = self.emb_dim * self.num_gc_layers
		else:
			raise NotImplementedError

		self.convs = torch.nn.ModuleList()
		self.bns = torch.nn.ModuleList()

		self.atom_encoder = AtomEncoder(emb_dim)
		self.bond_encoder = BondEncoder(emb_dim)
		self.edge_dim = edge_dim

		if convolution == "gin":
			print(f"Using GIN backbone for {num_gc_layers} layers")
			self.convolution = GINEConv

			for i in range(num_gc_layers):
				nn = Sequential(Linear(emb_dim, 2 * emb_dim), torch.nn.BatchNorm1d(2 * emb_dim), ReLU(),
								Linear(2 * emb_dim, emb_dim))
				conv = GINEConv(nn)
				bn = torch.nn.BatchNorm1d(emb_dim)
				self.convs.append(conv)
				self.bns.append(bn)
		else:
			raise NotImplementedError

		self.init_emb()

	def init_emb(self):
		"""
		Initializes the node embeddings.
		"""
		for m in self.modules():
			if isinstance(m, Linear):
				torch.nn.init.xavier_uniform_(m.weight.data)
				if m.bias is not None:
					m.bias.data.fill_(0.0)

	def forward(self, batch, x, edge_index, edge_attr, edge_weight=None):
		"""
		Performs forward pass through the encoder.

		Args:
			batch (Tensor): The batch tensor.
			x (Tensor): The node feature tensor.
			edge_index (LongTensor): The edge index tensor.
			edge_attr (Tensor): The edge attribute tensor.
			edge_weight (Tensor, optional): The edge weight tensor. Defaults to None.

		Returns:
			Tuple[Tensor, Tensor]: The graph embedding and node embedding tensors.

		"""
		# print(x, x.shape)
		x = self.atom_encoder(x.to(torch.int))
		edge_attr = self.bond_encoder(edge_attr.to(torch.int))
		# compute node embeddings using GNN
		xs = []
		for i in range(self.num_gc_layers):

			if edge_weight is None:
				edge_weight = torch.ones((edge_index.shape[1], 1)).to(x.device)

			if self.convolution == GINEConv:
				x = self.convs[i](x, edge_index, edge_attr, edge_weight)

			x = self.bns[i](x)
			if i == self.num_gc_layers - 1:
				# remove relu for the last layer
				x = F.dropout(x, self.drop_ratio, training=self.training)
			else:
				x = F.dropout(F.relu(x), self.drop_ratio, training=self.training)
			xs.append(x)

		# compute graph embedding using pooling
		if self.pooling_type == "standard":
			xpool = global_add_pool(x, batch)
			return normalize_l2(xpool), x

		elif self.pooling_type == "layerwise":
			xpool = [global_add_pool(x, batch) for x in xs]
			xpool = torch.cat(xpool, 1)
			if self.is_infograph:
				return xpool, torch.cat(xs, 1)
			else:
				return xpool, x
		else:
			raise NotImplementedError

	def get_embeddings(self, loader, device, is_rand_label=False, every=1, node_features=False):
		"""
		Computes embeddings for a given data loader.

		Args:
			loader (DataLoader): The data loader.
			device (torch.device): The device to perform computations on.
			is_rand_label (bool, optional): Whether to use random labels. Defaults to False.
			every (int, optional): The interval at which to compute embeddings. Defaults to 1.
			node_features (bool, optional): Whether to use node features. Defaults to False.

		Returns:
			Tuple[np.ndarray, np.ndarray]: The computed embeddings and labels.

		"""
		ret = []
		y = []
		with torch.no_grad():
			for i, data in enumerate(loader):
				if i % every != 0:
					continue

				if isinstance(data, list):
					data = data[0].to(device)

				data = data.to(device)
				batch, x, edge_index, edge_attr = data.batch, data.x, data.edge_index, data.edge_attr

				# Hard coding for now - should find a smarter way of doing this during evaluation
				x = x[:, 0].reshape(-1, 1)
				edge_attr = edge_attr[:, 0].reshape(-1, 1)

				if not node_features:
					x = torch.ones((x.shape[0], 1)).to(device)
					edge_attr = torch.ones((edge_attr.shape[0], 1)).to(device)

				edge_weight = data.edge_weight if hasattr(data, 'edge_weight') else None

				if x is None:
					x = torch.ones((batch.shape[0], 1)).to(device)

				# print(x) 
				# print(edge_index.shape)
				# print(torch.min(edge_index))
				# print(torch.max(edge_index))
				# print(edge_attr.shape)

				x, _ = self.forward(batch, x, edge_index, edge_attr, edge_weight)

				ret.append(x.cpu().numpy())

				try:
					if is_rand_label:
						y.append(data.rand_label.cpu().numpy())
					else:
						y.append(data.y.cpu().numpy())
				except AttributeError:
					y.append(torch.ones(x.shape[0]).to(torch.float))
		ret = np.concatenate(ret, 0)
		y = np.concatenate(y, 0)
		return ret, y




class NodeEncoder(torch.nn.Module):
	"""
	NodeEncoder is a module that performs node encoding in a graph neural network.

	Args:
		emb_dim (int): The dimensionality of the node embeddings.
		num_gc_layers (int): The number of graph convolutional layers.
		drop_ratio (float): The dropout ratio.
		pooling_type (str): The type of pooling to use for graph embedding.
		is_infograph (bool): Whether to use Infograph pooling.
		convolution (torch.nn.Module): The graph convolutional layer to use.

	Attributes:
		pooling_type (str): The type of pooling used for graph embedding.
		emb_dim (int): The dimensionality of the node embeddings.
		num_gc_layers (int): The number of graph convolutional layers.
		drop_ratio (float): The dropout ratio.
		is_infograph (bool): Whether to use Infograph pooling.
		out_node_dim (int): The output dimensionality of the node embeddings.
		out_graph_dim (int): The output dimensionality of the graph embeddings.
		convs (torch.nn.ModuleList): The list of graph convolutional layers.
		bns (torch.nn.ModuleList): The list of batch normalization layers.
		atom_encoder (AtomEncoder): The atom encoder module.
		bond_encoder (BondEncoder): The bond encoder module.

	Methods:
		init_emb(): Initializes the node embeddings.
		forward(batch, x, edge_index, edge_attr, edge_weight=None): Performs forward pass through the module.
		get_embeddings(loader, device, is_rand_label=False, every=1, node_features=False): Computes node embeddings.

	"""

	def __init__(self, emb_dim=300, num_gc_layers=5, drop_ratio=0.0,
				 pooling_type="standard", is_infograph=False, convolution=GINEConv):
		super(NodeEncoder, self).__init__()

		self.pooling_type = pooling_type
		self.emb_dim = emb_dim
		self.num_gc_layers = num_gc_layers
		self.drop_ratio = drop_ratio
		self.is_infograph = is_infograph

		self.out_node_dim = self.emb_dim
		if self.pooling_type == "standard":
			self.out_graph_dim = self.emb_dim
		elif self.pooling_type == "layerwise":
			self.out_graph_dim = self.emb_dim * self.num_gc_layers
		else:
			raise NotImplementedError

		self.convs = torch.nn.ModuleList()
		self.bns = torch.nn.ModuleList()

		self.atom_encoder = AtomEncoder(emb_dim)
		self.bond_encoder = BondEncoder(emb_dim)

		if convolution != GATv2Conv:
			for i in range(num_gc_layers):
				nn = Sequential(Linear(emb_dim, 2*emb_dim), torch.nn.BatchNorm1d(2*emb_dim), ReLU(), Linear(2*emb_dim, emb_dim))
				conv = convolution(nn)
				bn = torch.nn.BatchNorm1d(emb_dim)
				self.convs.append(conv)
				self.bns.append(bn)
		else:
			for i in range(num_gc_layers):
				conv = convolution(in_channels=self.emb_dim,
								   out_channels=self.out_node_dim,
								   heads=1)
				self.convs.append(conv)
				bn = torch.nn.BatchNorm1d(emb_dim)
				self.bns.append(bn)

		self.init_emb()

	def init_emb(self):
		"""
		Initializes the node embeddings.
		"""
		for m in self.modules():
			if isinstance(m, Linear):
				torch.nn.init.xavier_uniform_(m.weight.data)
				if m.bias is not None:
					m.bias.data.fill_(0.0)

	def forward(self, batch, x, edge_index, edge_attr, edge_weight=None):
		"""
		Performs forward pass through the module.

		Args:
			batch (torch.Tensor): The batch tensor.
			x (torch.Tensor): The node feature tensor.
			edge_index (torch.Tensor): The edge index tensor.
			edge_attr (torch.Tensor): The edge attribute tensor.
			edge_weight (torch.Tensor, optional): The edge weight tensor.

		Returns:
			torch.Tensor: The graph embedding tensor.
			torch.Tensor: The node embedding tensor.
		"""
		x = self.atom_encoder(x.to(torch.int))
		edge_attr = self.bond_encoder(edge_attr.to(torch.int))

		xs = []
		for i in range(self.num_gc_layers):
			x = self.convs[i](x, edge_index, edge_attr, edge_weight)
			x = self.bns[i](x)
			if i == self.num_gc_layers - 1:
				x = F.dropout(x, self.drop_ratio, training=self.training)
			else:
				x = F.dropout(F.relu(x), self.drop_ratio, training=self.training)
			xs.append(x)

		if self.pooling_type == "standard":
			xpool = global_add_pool(x, batch)
			return xpool, x
		elif self.pooling_type == "layerwise":
			xpool = [global_add_pool(x, batch) for x in xs]
			xpool = torch.cat(xpool, 1)
			if self.is_infograph:
				return xpool, torch.cat(xs, 1)
			else:
				return xpool, x
		else:
			raise NotImplementedError

	def get_embeddings(self, loader, device, is_rand_label=False, every=1, node_features=False):
		"""
		Computes node embeddings.

		Args:
			loader (torch.utils.data.DataLoader): The data loader.
			device (torch.device): The device to use for computation.
			is_rand_label (bool, optional): Whether to use random labels.
			every (int, optional): The interval for computing embeddings.
			node_features (bool, optional): Whether to use node features.

		Returns:
			numpy.ndarray: The computed node embeddings.
			numpy.ndarray: The corresponding labels.
		"""
		ret = []
		y = []
		with torch.no_grad():
			for i, data in enumerate(loader):
				if i % every != 0:
					continue

				if isinstance(data, list):
					data = data[0].to(device)

				data = data.to(device)
				batch, x, edge_index, edge_attr = data.batch, data.x, data.edge_index, data.edge_attr

				if not node_features:
					x = torch.ones_like(x)

				edge_weight = data.edge_weight if hasattr(data, 'edge_weight') else None

				if x is None:
					x = torch.ones((batch.shape[0], 1)).to(device)
				xpool, x = self.forward(batch.to(device), x.to(device), edge_index.to(device), edge_attr.to(device), edge_weight.to(device))

				ret.append(x.cpu().numpy())
				try:
					if is_rand_label:
						y.append(data.rand_label.cpu().numpy())
					else:
						y.append(data.y.cpu().numpy())
				except AttributeError:
					y.append(torch.ones(x.shape[0]).to(torch.float))
		ret = np.concatenate(ret, 0)
		y = np.concatenate(y, 0)
		return ret, y



import os
import networkx as nx
import torch
from tqdm import tqdm
from torch_geometric.data import InMemoryDataset, Data
from torch_geometric.utils.convert import to_networkx
from torch_geometric.io import read_npz
from torch.nn.functional import one_hot
import wget
import matplotlib.pyplot as plt
from ..utils import describe_one_dataset, vis_grid, ESWR
from ..utils import ESWR



def specific_from_networkx(graph):
    # Turns a graph into a pytorch geometric object
    # Mostly by unpacking dictionaries on nodes and edges
    # Here node labels are the target
    # One of these functions for each dataset ideally - they are unlikely to transfer across datasets
    node_labels = []
    node_attrs = []
    edge_indices = []

    # Collect node labels and attributes
    for n in list(graph.nodes(data=True)):
        # list(graph.nodes(data=True)) returns [(node_id1, {attribute dictionary}), (node_id2, ...), (node_id3, ...)]
        node_labels.append(n[1]["label"])
        node_attrs.append(n[1]["attrs"])

    # Collect edge indices and attributes
    for e in graph.edges(data=True):
        # graph.edges(data=True) is a generator producing (node_id1, node_id2, {attribute dictionary})
        edge_indices.append((e[0], e[1]))

        # uncomment for edge attributes:
        # edge_attrs.append(e[2]["attrs"]) 

    # Convert to PyTorch tensors
    node_labels = torch.stack(node_labels).flatten()

    # Specific to classification on nodes! Hard coding num classes as this happens on a per-graph basis
    # Per-graph classification should also be one-hot
    node_labels = one_hot(node_labels.to(int), num_classes = 7)

    node_attrs = torch.stack(node_attrs)
    edge_indices = torch.tensor(edge_indices, dtype=torch.long).t().contiguous()

    # Create PyG Data object
    # Can pass:
    # x:            node features, shape (n nodes x n features)
    # edge_index:   the list of edges in the graph, shape (2, n_edges). Entries edge_index[i, :] are [node_id1, node_id2].
    # edge_attr:    edge features, shape (n_edges, n_features), same order as edgelist
    # y:            targets. Graph regression shape (n_variables), graph classification (n_classes), node classification (n_nodes, n_classes), edge classification (n_edges, n_classes)
    data = Data(x=node_attrs, edge_index=edge_indices, edge_attr = None,  y=node_labels)

    return data

def download_example():
    # This function should download the data and process it into one big networkx graph

    # If node labels are the target, they should be included as an attribute of each node
    # In this case each node is (id, {"attrs":..., "label":...})
    # Probably the same is best for edge labels

    # url to data
    zip_url = "https://github.com/abojchevski/graph2gauss/raw/master/data/cora_ml.npz"

    # swap into directory
    start_dir = os.getcwd()
    os.chdir("bgd_files")

    # create directory if we haven't before
    if "example" not in os.listdir():
        print("Downloading CORA (as an example!)")
        os.mkdir("example")
        os.chdir("example")
    # Means that we've previously downloaded data
    else:
        os.chdir("example")

    if "cora_ml.npz" not in os.listdir():
        _ = wget.download(zip_url)
    edges = read_npz("cora_ml.npz")

    G = to_networkx(edges, to_undirected=True)

    node_classes = {n: int(edges.y[i].item()) for i, n in enumerate(list(G.nodes()))}
    node_attrs = {n: edges.x[i] for i, n in enumerate(list(G.nodes()))}

    for node in list(G.nodes()):
        class_tensor = torch.Tensor([node_classes[node]])
        node_attr = node_attrs[node]

        # If you have node features:
        G.nodes[node]["attrs"] = node_attr

        # If you have node labels:
        G.nodes[node]["label"] = class_tensor

        # If you don't have any node features:
        # G.nodes[node]["attrs"] = torch.Tensor([1])

    # No edge features for CORA, uncomment if you need
    # for edge in list(G.edges()):
    #     G.edges[edge]["attrs"] = torch.Tensor([1])
    #     G.edges[edge]["label"] = torch.Tensor([1])

    # This is generic cleaning stuff - we take the largest connected component and re-set node ids
    CGs = [G.subgraph(c) for c in nx.connected_components(G)]
    CGs = sorted(CGs, key=lambda x: x.number_of_nodes(), reverse=True)
    graph = CGs[0]
    graph = nx.convert_node_labels_to_integers(graph)

    # Important! Need to move back to start directory for other code
    os.chdir(start_dir)

    return graph

def get_example_dataset(num = 2000):
    # Get the cora graph
    cora_graph = download_example()

    # ESWR samples a load of graphs from a large input graph
    # Doesn't handle graph labels! See download_example for how to deal with this
    # Arguments are networkx.Graph, num graphs to sample, size doesn't currently do anything
    nx_graph_list = ESWR(cora_graph, num, 96)

    data_objects = [specific_from_networkx(graph) for graph in tqdm(nx_graph_list, desc = "Converting back to pyg graphs")]

    return  data_objects

class ExampleDataset(InMemoryDataset):
    # Documentation is essential! Without the sources listed I won't be able to include the dataset
    r"""
    Contributor: Alex O. Davies
    
    Contributor email: `alexander.davies@bristol.ac.uk`
    

    Academic citation graphs from the ML community, sampled from a large original graph using ESWR.
    The original graph is sourced from:

         `Yang, Zhilin, William Cohen, and Ruslan Salakhudinov. "Revisiting semi-supervised learning with graph embeddings." International conference on machine learning. PMLR, 2016.`

    The original data has one-hot bag-of-words over paper abstract as node features.

    The task is node classification for the category of each paper, one-hot encoded for seven categories.

     - Task: Node classification
     - Num node features: 2879
     - Num edge features: None
     - Num target values: 7
     - Target shape: N Nodes
     - Num graphs: Parameterised by `num`

    Args:
        root (str): Root directory where the dataset should be saved.
        stage (str): The stage of the dataset to load. One of "train", "val", "test". (default: :obj:`"train"`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. (default: :obj:`2000`).
    """
    def __init__(self, root, stage = "train", transform=None, pre_transform=None, pre_filter=None, num = 2000):
        self.num = num
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2}

        # Options are node-classification, node-regression, graph-classification, graph-regression, edge-regression, edge-classification
        # Graph-level tasks are preferred! (graph-classification and graph-regression)
        # edge-prediction is another option if you can't think of a good task
        self.task = "node-classification"

        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])


    @property
    def raw_file_names(self):
        # Replace with your saved raw file name
        return ['cora_ml.npz']

    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.

        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print(f"Cora files exist")
            return

        # Get a list of num pytorch_geometric.data.Data objects
        data_list = get_example_dataset(num=self.num)

        # You can iterate over the data objects if necessary:
        # ===================================================
        # new_data_list = []
        # for i, item in enumerate(data_list):
        #     n_nodes, n_edges = item.x.shape[0], item.edge_index.shape[1]

        #     data = Data(x = item.x,
        #                 edge_index=item.edge_index,
        #                 # Here we don't have any edge features so we use just 1s
        #                 edge_attr=None,
        #                 y = item.y)

        #     new_data_list.append(data)
        # data_list = new_data_list
        # ===================================================

        # Torch geometric stuff
        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        # Save data
        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


# Run with python -m bgd/example_dataset.py
# Run this to check that everything works!
if __name__ == "__main__":
    # Please set the last part of the path to your dataset name!
    dataset = ExampleDataset(os.getcwd()+'/bgd_files/'+'example', stage = "train")
    describe_one_dataset(dataset)
    vis_grid(dataset[:16], os.getcwd()+"/bgd_files/example/train.png")

    dataset = ExampleDataset(os.getcwd()+'/bgd_files/'+'example', stage = "val")
    describe_one_dataset(dataset)
    vis_grid(dataset[:16], os.getcwd()+"/bgd_files/example/val.png")

    dataset = ExampleDataset(os.getcwd()+'/bgd_files/'+'example', stage = "test")
    describe_one_dataset(dataset)
    vis_grid(dataset[:16], os.getcwd()+"/bgd_files/example/test.png")


# from .loaders import get_test_datasets, get_val_datasets, get_train_datasets, get_all_datasets
# from .cora_dataset import CoraDataset
# from .road_dataset import RoadDataset
# from .neural_dataset import NeuralDataset
# from .facebook_dataset import FacebookDataset
# from .reddit_dataset import RedditDataset
# from .ego_dataset import EgoDataset
# from .random_dataset import RandomDataset
# from .community_dataset import CommunityDataset
# from .tree_dataset import TreeDataset

from bgd import real
from bgd import synthetic
from bgd import loaders

from .loaders import *
from .real import *
from .synthetic import *

_real_datasets = ['CoraDataset', 'RoadDataset', 'NeuralDataset',
                'FacebookDataset', 'RedditDataset', 'EgoDataset',
                'FromOGBDataset', 'from_ogb_dataset',
                'FromTUDataset', 'from_tu_dataset']

_synthetic_datasets = ['RandomDataset', 'CommunityDataset', 'TreeDataset']


_loaders = ['get_test_datasets', 'get_val_datasets', 'get_train_datasets', 'get_datasets',
            'get_all_datasets', 'get_graph_task_datasets', 'get_edge_task_datasets',
            'get_node_task_datasets', 'get_graph_classification_datasets', 'get_graph_regression_datasets']

_modules = ['real', 'synthetic', 'loaders']

__all__ = _real_datasets + _synthetic_datasets + _loaders + _modules


import os
import torch
from torch.nn.functional import one_hot
import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
from torch_geometric.data import InMemoryDataset, Data
from tqdm import tqdm
from ..utils import describe_one_dataset

from torch_geometric.datasets import TUDataset

# from ogb.utils.features import get_atom_feature_dims, get_bond_feature_dims
# from ogb.graphproppred import PygGraphPropPredDataset


# full_atom_feature_dims = get_atom_feature_dims()
# full_bond_feature_dims = get_bond_feature_dims()


# def to_onehot_atoms(x):
#     one_hot_tensors = []
#     for i, num_values in enumerate(full_atom_feature_dims):
#         one_hot = torch.nn.functional.one_hot(x[:, i], num_classes=num_values)
#         one_hot_tensors.append(one_hot)

#     return torch.cat(one_hot_tensors, dim=1)

# def to_onehot_bonds(x):
#     one_hot_tensors = []
#     for i, num_values in enumerate(full_bond_feature_dims):
#         one_hot = torch.nn.functional.one_hot(x[:, i], num_classes=num_values)
#         one_hot_tensors.append(one_hot)

#     return torch.cat(one_hot_tensors, dim=1)

class FromTUDataset(InMemoryDataset):
    r"""
    Contributor: Alex O. Davies (minimal alterations to existing code from PyG)
    
    Contributor email: `alexander.davies@bristol.ac.uk`

    Returns a `torch_geometric.data.InMemoryDataset` for each TUDataset.
    This allows standard dataset operations like concatenation with other datasets.

    The datasets were originally collected in this paper:

        `Morris, Christopher, et al. "TUDataset: A collection of benchmark datasets for learning with graphs." (2020).`

    Args:
        root (str): Root directory where the dataset should be saved.
        ogb_dataset (list): a TUDataset to be converted back to `InMemoryDataset`.
        stage (str): The stage of the dataset to load. One of "train", "val", "test", "None". If None, returns the whole original dataset. Otherwise returns one of a (80,10,10) train/val/test split. (default: :obj:`None`)
        transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in an :obj:`torch_geometric.data.Data` object and returns a transformed version. The data object will be transformed before being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an :obj:`torch_geometric.data.Data` object and returns a boolean value, indicating whether the data object should be included in the final dataset. (default: :obj:`None`)
        num (int): The number of samples to take from the original dataset. -1 takes all available samples for that stage. Ignored if stage is not None. (default: :obj:`-1`).

    **STATS:**

    .. list-table::
        :widths: 20 10 10 10 10 10
        :header-rows: 1

        * - Name
          - #graphs
          - #nodes
          - #edges
          - #features
          - #classes
        * - MUTAG
          - 188
          - ~17.9
          - ~39.6
          - 7
          - 2
        * - ENZYMES
          - 600
          - ~32.6
          - ~124.3
          - 3
          - 6
        * - PROTEINS
          - 1,113
          - ~39.1
          - ~145.6
          - 3
          - 2
        * - COLLAB
          - 5,000
          - ~74.5
          - ~4914.4
          - 0
          - 3
        * - IMDB-BINARY
          - 1,000
          - ~19.8
          - ~193.1
          - 0
          - 2
        * - REDDIT-BINARY
          - 2,000
          - ~429.6
          - ~995.5
          - 0
          - 2
        * - ...
          -
          -
          -
          -
          -
    """
    def __init__(self, root, ogb_dataset, stage = "train", num = -1, transform=None, pre_transform=None, pre_filter=None):
        self.ogb_dataset = ogb_dataset
        self.stage = stage
        self.stage_to_index = {"train":0,
                               "val":1,
                               "test":2,
                               "train-adgcl":3}
        self.num = num
        self.task = "graph-classification"
        print(f"Converting OGB stage {self.stage}")
        super().__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[self.stage_to_index[self.stage]])

    @property
    def raw_file_names(self):
        return ['dummy.csv']

    @property
    def processed_file_names(self):
        return ['train.pt',
                'val.pt',
                'test.pt']


    def process(self):
        # Read data into huge `Data` list.
        if os.path.isfile(self.processed_paths[self.stage_to_index[self.stage]]):
            print(f"\nTU files exist at {self.processed_paths[self.stage_to_index[self.stage]]}")
            return
        data_list = self.ogb_dataset

        num_classes = data_list.num_classes

        num_samples = len(data_list)
        if self.stage is None:
            if num_samples < self.num:
                keep_n = num_samples
            else:
                keep_n = self.num
        else:
            if self.stage == "train":
                data_list = data_list[:int(0.8 * num_samples)]
            elif self.stage == "val":
                data_list = data_list[int(0.8 * num_samples):int(0.9 * num_samples)]
            elif self.stage == "test":
                data_list = data_list[int(0.9 * num_samples):]

            keep_n = len(data_list)

        new_data_list = []
        for i, item in enumerate(data_list[:keep_n]):

            data = Data(x = item.x, 
                        edge_index=item.edge_index,
                        edge_attr= item.edge_attr, 
                        y = one_hot(item.y, num_classes = num_classes))
            
            new_data_list.append(data)
        data_list = new_data_list

        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[self.stage_to_index[self.stage]])


def from_tu_dataset(root,  stage="train", num=-1):
    """
    Load a TU dataset and convert it to the Big Graph Dataset format.

    Args:
        name (str): The name of the TU dataset. ("MUTAG", "ENZYMES", "PROTEINS", "COLLAB", "IMDB-BINARY", "REDDIT-BINARY")
        stage (str, optional): The stage of the dataset to load (e.g., "train", "valid", "test", None). Defaults to None, which returns the whole dataset. Otherwise returns one of a (80/10/10) train/val/test split.
        num (int, optional): The number of samples to load. Set to -1 to load all samples. Defaults to -1. Ignored if stage is not None.

    Returns:
        FromOGBDataset: The converted dataset in the Big Graph Dataset format.
    """
    dataset = TUDataset(root, root.split('/')[-1])
    return FromTUDataset(root, dataset, stage = stage, num = num)

if __name__ == "__main__":
    mutag = from_tu_dataset("PROTEINS", stage = "train")
    describe_one_dataset(mutag)


import math

import torch
import torch.nn.functional as F
from torch import nn


def log_sum_exp(x, axis=None):
	"""Log sum exp function
	Args:
		x: Input.
		axis: Axis over which to perform sum.
	Returns:
		torch.Tensor: log sum exp
	"""
	x_max = torch.max(x, axis)[0]
	y = torch.log((torch.exp(x - x_max)).sum(axis)) + x_max
	return y
def raise_measure_error(measure):
	supported_measures = ['GAN', 'JSD', 'X2', 'KL', 'RKL', 'DV', 'H2', 'W1']
	raise NotImplementedError(
		'Measure `{}` not supported. Supported: {}'.format(measure,
														   supported_measures))

def get_positive_expectation(p_samples, measure, average=True):
	"""Computes the positive part of a divergence / difference.
	Args:
		p_samples: Positive samples.
		measure: Measure to compute for.
		average: Average the result over samples.
	Returns:
		torch.Tensor
	"""
	log_2 = math.log(2.)

	if measure == 'GAN':
		Ep = - F.softplus(-p_samples)
	elif measure == 'JSD':
		Ep = log_2 - F.softplus(- p_samples)
	elif measure == 'X2':
		Ep = p_samples ** 2
	elif measure == 'KL':
		Ep = p_samples + 1.
	elif measure == 'RKL':
		Ep = -torch.exp(-p_samples)
	elif measure == 'DV':
		Ep = p_samples
	elif measure == 'H2':
		Ep = 1. - torch.exp(-p_samples)
	elif measure == 'W1':
		Ep = p_samples
	else:
		raise_measure_error(measure)

	if average:
		return Ep.mean()
	else:
		return Ep


def get_negative_expectation(q_samples, measure, average=True):
	"""Computes the negative part of a divergence / difference.
	Args:
		q_samples: Negative samples.
		measure: Measure to compute for.
		average: Average the result over samples.
	Returns:
		torch.Tensor
	"""
	log_2 = math.log(2.)

	if measure == 'GAN':
		Eq = F.softplus(-q_samples) + q_samples
	elif measure == 'JSD':
		Eq = F.softplus(-q_samples) + q_samples - log_2
	elif measure == 'X2':
		Eq = -0.5 * ((torch.sqrt(q_samples ** 2) + 1.) ** 2)
	elif measure == 'KL':
		Eq = torch.exp(q_samples)
	elif measure == 'RKL':
		Eq = q_samples - 1.
	elif measure == 'DV':
		Eq = log_sum_exp(q_samples, 0) - math.log(q_samples.size(0))
	elif measure == 'H2':
		Eq = torch.exp(q_samples) - 1.
	elif measure == 'W1':
		Eq = q_samples
	else:
		raise_measure_error(measure)

	if average:
		return Eq.mean()
	else:
		return Eq

def local_global_loss_(l_enc, g_enc, batch, measure):
	'''
	Args:
		l: Local feature map.
		g: Global features.
		measure: Type of f-divergence. For use with mode `fd`
		mode: Loss mode. Fenchel-dual `fd`, NCE `nce`, or Donsker-Vadadhan `dv`.
	Returns:
		torch.Tensor: Loss.
	'''
	num_graphs = g_enc.shape[0]
	num_nodes = l_enc.shape[0]

	pos_mask = torch.zeros((num_nodes, num_graphs)).cuda()
	neg_mask = torch.ones((num_nodes, num_graphs)).cuda()
	for nodeidx, graphidx in enumerate(batch):
		pos_mask[nodeidx][graphidx] = 1.
		neg_mask[nodeidx][graphidx] = 0.

	res = torch.mm(l_enc, g_enc.t())

	E_pos = get_positive_expectation(res * pos_mask, measure, average=False).sum()
	E_pos = E_pos / num_nodes
	E_neg = get_negative_expectation(res * neg_mask, measure, average=False).sum()
	E_neg = E_neg / (num_nodes * (num_graphs - 1))

	return E_neg - E_pos


class PriorDiscriminator(nn.Module):
	def __init__(self, input_dim):
		super().__init__()
		self.l0 = nn.Linear(input_dim, input_dim)
		self.l1 = nn.Linear(input_dim, input_dim)
		self.l2 = nn.Linear(input_dim, 1)

	def forward(self, x):
		h = F.relu(self.l0(x))
		h = F.relu(self.l1(h))
		return torch.sigmoid(self.l2(h))


class FF(nn.Module):
	def __init__(self, input_dim):
		super().__init__()
		self.block = nn.Sequential(
			nn.Linear(input_dim, input_dim),
			nn.ReLU(),
			nn.Linear(input_dim, input_dim),
			nn.ReLU(),
			nn.Linear(input_dim, input_dim),
			nn.ReLU()
		)
		self.linear_shortcut = nn.Linear(input_dim, input_dim)

	def forward(self, x):
		return self.block(x) + self.linear_shortcut(x)


class InfoGraph(nn.Module):
	def __init__(self, encoder, alpha=0.5, beta=1., gamma=.1, prior=True):
		super(InfoGraph, self).__init__()

		self.alpha = alpha
		self.beta = beta
		self.gamma = gamma
		self.prior = prior


		self.encoder = encoder
		self.embedding_dim = self.encoder.out_graph_dim

		self.local_d = FF(self.embedding_dim)
		self.global_d = FF(self.embedding_dim)

		if self.prior:
			self.prior_d = PriorDiscriminator(self.embedding_dim)

		self.init_emb()

	def init_emb(self):
		for m in self.modules():
			if isinstance(m, nn.Linear):
				torch.nn.init.xavier_uniform_(m.weight.data)
				if m.bias is not None:
					m.bias.data.fill_(0.0)

	def forward(self, batch, x, edge_index, edge_attr=None, edge_weight=None):
		# batch_size = data.num_graphs

		y, M = self.encoder(batch, x, edge_index, edge_attr, edge_weight)

		g_enc = self.global_d(y)
		l_enc = self.local_d(M)

		mode = 'fd'
		measure = 'JSD'
		local_global_loss = local_global_loss_(l_enc, g_enc, batch, measure)

		if self.prior:
			prior = torch.rand_like(y)
			term_a = torch.log(self.prior_d(prior)).mean()
			term_b = torch.log(1.0 - self.prior_d(y)).mean()
			PRIOR = - (term_a + term_b) * self.gamma
		else:
			PRIOR = 0

		return local_global_loss + PRIOR


import logging
import random

import numpy as np
import torch
import yaml
import os
import wget

from unsupervised.encoder import Encoder
from unsupervised.learning import GInfoMinMax
from torch_geometric.loader import DataLoader

from sklearn.metrics.pairwise import cosine_similarity
from umap import UMAP
from sklearn.decomposition import PCA
import matplotlib
import matplotlib.pyplot as plt
from sklearn.preprocessing import normalize, StandardScaler

from tqdm import tqdm


def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    np.random.seed(seed)
    random.seed(seed)

def get_emb_y(loader, encoder, device, dtype='numpy', is_rand_label=False, every = 1, node_features = False):
    with torch.no_grad():
        x, y = encoder.get_embeddings(loader, device, is_rand_label, every = every, node_features = node_features)
    if dtype == 'numpy':
        return x,y
    elif dtype == 'torch':
        return torch.from_numpy(x).to(device), torch.from_numpy(y).to(device)
    else:
        raise NotImplementedError

def annotate_heatmap(im, data=None, valfmt="{x:.2f}",
                     textcolors=("black", "white"),
                     threshold=None, **textkw):
    """
    A function to annotate a heatmap.

    Parameters
    ----------
    im
        The AxesImage to be labeled.
    data
        Data used to annotate.  If None, the image's data is used.  Optional.
    valfmt
        The format of the annotations inside the heatmap.  This should either
        use the string format method, e.g. "$ {x:.2f}", or be a
        `matplotlib.ticker.Formatter`.  Optional.
    textcolors
        A pair of colors.  The first is used for values below a threshold,
        the second for those above.  Optional.
    threshold
        Value in data units according to which the colors from textcolors are
        applied.  If None (the default) uses the middle of the colormap as
        separation.  Optional.
    **kwargs
        All other arguments are forwarded to each call to `text` used to create
        the text labels.
    """

    if data is None:
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        # For data between -1 and 1, the middle is 0
        threshold = im.norm(0)

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts

class GeneralEmbeddingEvaluation:
    """
    Class for evaluating embeddings and visualizing the results.

    Methods:
        __init__(): Initializes the GeneralEmbeddingEvaluation object.
        embedding_evaluation(encoder, train_loaders, names): Performs embedding evaluation using the given encoder, train loaders, and names.
        get_embeddings(encoder, loaders, names): Retrieves the embeddings from the encoder and loaders.
        vis(all_embeddings, separate_embeddings, names): Visualizes the embeddings using UMAP and PCA projections.
        centroid_similarities(embeddings, names): Calculates the pairwise similarities between the centroids of the embeddings.
    """
    def __init__(self):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def embedding_evaluation(self, encoder, train_loaders, names):
        """
        Evaluate the embeddings generated by the encoder.

        Args:
            encoder: The encoder model used to generate the embeddings.
            train_loaders: A list of data loaders for the data.
            names: A list of names corresponding to the data loaders.

        Returns:
            None
        """
        train_all_embeddings, train_separate_embeddings = self.get_embeddings(encoder, train_loaders, names)
        self.centroid_similarities(train_separate_embeddings, names)
        # self.vis(train_all_embeddings, train_separate_embeddings, names)

    def get_embeddings(self, encoder, loaders, names):
        """
        Get embeddings for the given encoder and loaders.

        Args:
            encoder: The encoder model.
            loaders: A list of data loaders.
            names: A list of names corresponding to the loaders.

        Returns:
            A tuple containing the concatenated embeddings of all loaders and a list of separate embeddings for each loader.
        """
        encoder.eval()
        all_embeddings = None
        separate_embeddings = []
        pbar = tqdm(loaders, desc="Getting embeddings")
        for i, loader in enumerate(pbar):
            pbar.set_description(names[i])
            train_emb, train_y = get_emb_y(loader, encoder, self.device, is_rand_label=False, every=1, node_features=False)

            separate_embeddings.append(train_emb)
            if all_embeddings is None:
                all_embeddings = train_emb
            else:
                all_embeddings = np.concatenate((all_embeddings, train_emb))

        scaler = StandardScaler().fit(all_embeddings)
        for embedding in separate_embeddings:
            embedding = scaler.transform(embedding)
            embedding = normalize(embedding)

        all_embeddings = scaler.transform(all_embeddings)
        all_embeddings = normalize(all_embeddings)

        return all_embeddings, separate_embeddings

    def vis(self, all_embeddings, separate_embeddings, names):
        """
        Visualizes the embeddings using UMAP and PCA projections.

        Args:
            all_embeddings (numpy.ndarray): The combined embeddings of all graphs.
            separate_embeddings (list): A list of separate embeddings for each graph.
            names (list): A list of names corresponding to each graph.

        Returns:
            None
        """
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 8))

        cmap = plt.get_cmap('viridis')
        unique_categories = np.unique(names)
        colors = cmap(np.linspace(0, 1, len(unique_categories)))
        color_dict = {category: color for category, color in zip(unique_categories, colors)}

        cmap = plt.get_cmap('autumn')
        unique_categories = np.unique(names)
        colors = cmap(np.linspace(0, 1, len(unique_categories)))
        mol_color_dict = {category: color for category, color in zip(unique_categories, colors)}

        embedder = UMAP(n_components=2, n_jobs=4, n_neighbors = 30).fit(all_embeddings)

        for i, emb in enumerate(separate_embeddings):
            proj = embedder.transform(emb)
            name = names[i]
            if "ogb" in name:
                plot_marker = "^"
                color = mol_color_dict[name]
            else:
                plot_marker = "x"
                color = color_dict[name]

            ax1.scatter(proj[:, 0], proj[:, 1],
                        alpha= 1 - proj.shape[0] / all_embeddings.shape[0], s = 5,
                        label=f"{names[i]}", # - {proj.shape[0]} graphs",
                        c = color, marker = plot_marker)

        ax1.set_title("UMAP Embedding")

        embedder = PCA(n_components=2).fit(all_embeddings)


        for i, emb in enumerate(separate_embeddings):
            proj = embedder.transform(emb)
            name = names[i]
            if "ogb" in name:
                plot_marker = "^"
                color = mol_color_dict[name]
            else:
                plot_marker = "x"
                color = color_dict[name]

            ax2.scatter(proj[:, 0], proj[:, 1],
                        alpha= 1 - proj.shape[0] / all_embeddings.shape[0], s = 5,
                        c = color, marker = plot_marker)
            
                # Get the legend handles and labels

        ax2.set_title("PCA Projection")

        handles, labels = ax2.get_legend_handles_labels()



        # Create a new legend with increased size for scatter points
        new_handles = []
        for ihandle, handle in enumerate(handles):
            new_handle = plt.Line2D([0], [0],
             marker="^" if "ogb" in labels[ihandle] else "o", color='w',
               markerfacecolor=handle.get_facecolor()[0], markersize=3000)
            new_handles.append(new_handle)

        fig.legend(handles = new_handles, labels = labels,
                    bbox_transform = fig.transFigure,
                     loc='lower center',
                     ncol = 5, frameon = False, bbox_to_anchor = (0.5, 0.))
        
        plt.tight_layout()
        plt.subplots_adjust(bottom=0.2)
        plt.savefig("outputs/embedding.png")
        plt.close()


    def centroid_similarities(self, embeddings, names):
        """
        Calculate centroid similarities for a given set of embeddings and names.

        Parameters:
        embeddings (list of numpy arrays): List of embeddings, where each embedding is a numpy array.
        names (list of str): List of names corresponding to the embeddings.

        Returns:
        None

        This method calculates the centroid similarities for a given set of embeddings. It first calculates the centroid
        for each embedding by taking the mean along the axis 0. Then, it calculates the pairwise similarities between
        the centroids using cosine similarity. Finally, it visualizes the pairwise similarities as a heatmap and saves
        the plot as "outputs/pairwise-similarity.png".
        """
        embed_dim = embeddings[0].shape[1]
        centroids = np.zeros((len(embeddings), embed_dim))

        for i, embedding in enumerate(embeddings):
            centroids[i, :] = np.mean(embedding, axis=0)

        pairwise_similarities = cosine_similarity(centroids)
        print(pairwise_similarities)

        fig, ax = plt.subplots(figsize=(7, 6))

        im = ax.imshow(pairwise_similarities, cmap="binary", vmin=-1, vmax=1)

        ax.set_xticks(np.arange(len(names)), labels=names)
        ax.set_yticks(np.arange(len(names)), labels=names)

        annotate_heatmap(im, valfmt="{x:.3f}")

        plt.savefig("outputs/pairwise-similarity.png")

def get_top_model():
    checkpoint_url = "https://github.com/neutralpronoun/general-gcl/raw/main/outputs/all-100.pt"
    config_url = "https://github.com/neutralpronoun/general-gcl/raw/main/outputs/all-100.yaml"

    if "bgd_models" not in os.listdir():
        os.mkdir("bgd_models")
    os.chdir("bgd_models")

    if "all-100.pt" not in os.listdir():
        checkpoint = wget.download(checkpoint_url)
    if "all-100.yaml" not in os.listdir():
        config = wget.download(config_url)

    os.chdir("../")



def compute_top_scores(datasets, names):
    """
    Computes the top scores for graph structures using the ToP encoder.

    ToP scores use a pre-trained ToP model to compute the similarity between graphs across datasets.

    This function will also produce embedding visualisations using GeneralEmbeddingEvaluation.

    Args:
        datasets (list): A list of datasets containing graph structures.
        names (list): A list of names corresponding to each dataset.

    Returns:
        None
    """
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S')
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    # setup_seed(args.seed)
    get_top_model()
    checkpoint = "all-100.pt"

    checkpoint_path = f"bgd_models/{checkpoint}"
    cfg_name = checkpoint.split('.')[0] + ".yaml"
    config_path = f"bgd_models/{cfg_name}"

    with open(config_path, 'r') as stream:
        try:
            # Converts yaml document to python object
            wandb_cfg = yaml.safe_load(stream)

            # Printing dictionary
            print(wandb_cfg)
        except yaml.YAMLError as e:
            print(e)

    args = wandb_cfg

    # Get datasets
    train_loaders = [DataLoader(data, batch_size=128) for data in datasets]
    
    model = GInfoMinMax(
        Encoder(emb_dim=args["emb_dim"]["value"],
                 num_gc_layers=args["num_gc_layers"]["value"], 
                 drop_ratio=args["drop_ratio"]["value"],
                pooling_type="standard"),
        proj_hidden_dim=args["emb_dim"]["value"]).to(device)

    model_dict = torch.load(checkpoint_path, map_location=torch.device('cpu'))
    model.load_state_dict(model_dict['encoder_state_dict'])


    # Get embeddings
    general_ee = GeneralEmbeddingEvaluation()
    model.eval()

    general_ee.embedding_evaluation(model.encoder, train_loaders, names)

            
